from report_ranger import validation
from report_ranger.utils.jinja_helpers import log_jinja2_error
from report_ranger.utils.mdread import markdown_from_file, process_template
from report_ranger.output_formatter.outputformatter import oxfordcomma
import os
import logging
import copy
import datetime
from collections import Counter
from num2words import num2words
import jinja2
import traceback
import cerberus
from report_ranger.helpers import make_list

log = logging.getLogger(__name__)


def vuln_from_file(file_loc, riskassessment, env=None, vulnerability_validation={}, scope=None, watcher=None):
    ''' Get the vulnerability markdown from a file, process the headers and make sure they're there. Return a Vulnerability object.'''
    log.debug(f"Loading vulnerability file {file_loc}")
    try:
        headers, vulnmarkdown = markdown_from_file(file_loc, env=env, watcher=watcher)
    except:
        log.warn(f"Could not read vulnerability file {file_loc}")

    if 'name' in headers:
        if not isinstance(headers['name'], str):
            log.warn(f"Vulnerability name not a string for file {file_loc}")
            name = file_loc
            headers['name'] = file_loc
        else:
            name = headers['name']
    elif 'Name' in headers:  # This is for legacy reasons, it used to be case insensitive so we still accept "Name"
        name = headers['Name']
        headers['name'] = headers['Name']
    else:
        log.warn("Vulnerability name not in headers for file {}".format(file_loc))
        name = file_loc
        headers['name'] = file_loc

    if scope != None:
        headers['scope'] = scope

    validation.validate_headers(
        vulnerability_validation, headers, validation.default_vulnerability_validation, name=name)

    # Generate the new env variables they can play with
    headers['cwd'] = os.path.dirname(os.path.join(
        os.path.curdir, file_loc))  # for including screenshots

    return Vulnerability(name, headers=headers, markdown=vulnmarkdown, riskassessment=riskassessment, filename=file_loc)


class VulnerabilityList:
    """Reads a directory for vulnerability definition files, arranges them, and stores the vulnerabilities as objects.

    The class is usually referenced inside the markdown file itself to display the vulnerability information. It keeps the following attributes:

    - scopes : a List of Scope objects corresponding to the vulnerability scopes
    """

    def __init__(self):
        self.scopes = []

    def add_from_dir(self, vulndir, env, riskassessment, vulnerability_validation={}, watcher=None):
        # Make sure it's a directory
        if not os.path.isdir(vulndir):
            raise FileNotFoundError("Directory '{0}' doesn't exist. Skipping import of vulnerabilities.".format(
                vulndir))

        log.debug("Reading vulnerabilities")
        vulnerabilities = []

        scopeintroductions = {}
        log.debug(f"Vulnerability add_from_dir WATCHER: {watcher}")

        # Get the vulnerability writeups from the vulndir
        for f in os.listdir(vulndir):
            if watcher:
                watcher.add_path(vulndir)
            # Scopes can be split into directories within the vulndir
            if os.path.isdir(os.path.join(vulndir, f)):
                if watcher:
                    watcher.add_path(os.path.join(vulndir, f))
                log.debug(
                    "* New vulnerability scope in directory '{}'".format((os.path.join(vulndir, f))))
                vdir = os.path.join(vulndir, f)
                for v in os.listdir(vdir):
                    if v == "template.md" or v == "template.md.rr":
                        continue
                    if v == "introduction.md" or v == "introduction.md.rr":
                        try:
                            scopeintroductions[f] = markdown_from_file(
                                os.path.join(vdir, v), env=env, watcher=watcher)
                        except:
                            log.warn(f"Could not read introduction file {v} in vulnerability directory {vdir}")
                        continue
                    fn, ext = os.path.splitext(v)
                    if ext == ".md" or ext == ".rr":  # Only process markdown files
                        filename = os.path.join(vdir, v)
                        log.debug(
                            "  - Reading vulnerability file '{}'".format(filename))
                        try:
                            vuln = vuln_from_file(
                                filename, riskassessment=riskassessment, env=env, vulnerability_validation=vulnerability_validation, scope=f, watcher=watcher)
                            vulnerabilities += [vuln]
                        except validation.ValidationError as e:
                            pass  # ValidationError already raises logs
                        except Exception as e:
                            log.error(
                                "Received exception when processing markdown for vulnerability {}: {}".format(filename, e.args))
                            log.error(
                                "Removing markdown for vulnerability {}".format(self.name))
                            self.markdown = ""
            else:  # It's a file
                if f == "template.md" or f == "template.md.rr":
                    continue
                if f == "introduction.md" or f == "introduction.md.rr":
                    try:
                        scopeintroduction = markdown_from_file(
                            os.path.join(vulndir, f), env=env, watcher=watcher)
                        if "scope" in scopeintroduction[0]:
                            # There's a scope defined in the headers of the introduction
                            scopeintroductions[f] = scopeintroduction
                        else:
                            scopeintroductions[''] = scopeintroduction
                    except:
                        log.warn(f"Could not read introduction file {v} in vulnerability directory {vdir}")
                    continue
                    
                log.debug(
                    "* Reading vulnerability file '{}'".format(os.path.join(vulndir, f)))
                fn, ext = os.path.splitext(f)
                if ext == ".md" or ext == ".rr":  # Only process markdown files
                    try:
                        vulnerabilities += [vuln_from_file(
                            os.path.join(vulndir, f), riskassessment=riskassessment, env=env, watcher=watcher)]
                    except validation.ValidationError as e:
                        pass  # ValidationError already raises logs
                    except Exception as e:
                        log.error(
                            "Received exception when processing markdown for vulnerability {}: {}".format(filename, e.args))
                        log.error(
                            "Removing markdown for vulnerability {}".format(self.name))
                        self.markdown = ""

        # Sort the vulnerabilities into scopes
        scopes = {}

        # This is a marker for the highest risk
        highestrisk = riskassessment.risks[0]
        nhighestrisk = 0

        suggest_rc = set()
        suggest_ar = set()

        # Suggest high level findings
        for vuln in vulnerabilities:
            if 'suggest_rc' in vuln.headers:
                suggest_rc.update(vuln.headers['suggest_rc'])
            if 'suggest_ar' in vuln.headers:
                suggest_ar.update(vuln.headers['suggest_ar'])

        for rc in suggest_rc:
            log.info("Suggested root cause: {}".format(rc))
        for ar in suggest_ar:
            log.info("Suggested additional recommendation: {}".format(ar))

        for i in vulnerabilities:
            if i.nrisk > nhighestrisk:  # We have a new highest risk!
                nhighestrisk = i.nrisk
                highestrisk = i.risk
            if not i.scope in scopes:  # New scope
                scopes[i.scope] = Scope(i.scope)
            scopes[i.scope].add_vulnerability(i)

        for scope, introduction in scopeintroductions.items():
            if scope not in scopes:
                scopes[scope] = Scope(scope)
            scopes[scope].set_introduction(introduction[0], introduction[1])

        # Add the highest risk to the variables
        self.highestrisk = highestrisk
        env.set_variable('highestrisk', highestrisk)
        log.debug("Highest risk of any vuln is {}".format(highestrisk))

        self.scopes = list(scopes.values())
        self.scopes.sort()

        # Assign references
        ref = 1
        for scope in self.scopes:
            for vulnerability in scope.vulnerabilities:
                vulnerability.ref = ref
                ref += 1

    def get_scope_list(self, validation=None, date=None):
        scopelist = {}
        if date:
            updatedvlist = self.updated(date)
            scopes = updatedvlist.scopes
        else:
            scopes = self.scopes
        for scope in scopes:
            scopelist[scope.name] = scope.get_list(validation)
        return scopelist

    def get_ref(self, vulnerability_name):
        """ Get the reference of a vulnerability by name. Usually called from a markdown file. """
        for scope in self.scopes:
            ref = scope.get_ref(vulnerability_name)
            if ref:
                return ref
        return ""

    def updated(self, date=datetime.date.today()):
        # Alias for string today
        if date == "today" or date == "now":
            date = datetime.date.today()

        newlist = VulnerabilityList()
        for scope in self.scopes:
            newlist.scopes.append(scope.updated(date))
        newlist.scopes.sort()
        return newlist

    def nvulns(self, riskrating=""):
        """ Get the number of vulnerabilities with the given risk rating """
        return sum(scope.nvulns(riskrating) for scope in self.scopes)

    def nriskdict(self):
        """ Get a dictionary with risk:nrisks for all vulnerabilities """
        nvulnlist = {}
        for scope in self.scopes:
            nvulnlist.update(scope.nriskdict())
        return nvulnlist

    def nrisklist(self):
        return oxfordcomma(["{} ({}) {} risk".format(num2words(nrisk), nrisk, risk)
                            for risk, nrisk in self.nriskdict().items()])

    def generate_markdown(self, env):
        for scope in self.scopes:
            scope.generate_markdown(env)


class Scope:
    """ Holds a vulnerability scope, a container for vulnerabilities.

    Has the following attributes:

    name : The name of the scope as a string
    vulnerabilities : A List of Vulnerability objects, corresponding to the vulnerabilities in this scope.
                      Should always be sorted by severity and weight.
    highestrisk : The highest risk in this scope as a string
    sectionweight : The weight as a number used for ordering the scopes in a VulnerabilityList (highest weight first).
                    Called sectionweight to have consistency with sections.
    introduction_headers : The environment headers of an introduction, used for Jinja templating the markdown
    introduction : The introduction markdown
    """

    def __init__(self, name, sectionweight=0):
        self.name = name
        self.vulnerabilities = []
        self.highestrisk = 0
        self.sectionweight = sectionweight
        self.introduction_headers = {}
        self.introduction = ""

        # Process the introduction section weight if it's there
        if "sectionweight" in self.introduction_headers and self.introduction_headers["sectionweight"] > self.sectionweight:
            log.info(f"Setting sectionweight for {name} to {self.introduction_headers['sectionweight']} based on introduction header")
            self.sectionweight = self.introduction_headers["sectionweight"]

    def __lt__(self, other):
        if other.sectionweight == self.sectionweight:
            return other.highestrisk < self.highestrisk
        else:
            return other.sectionweight < self.sectionweight

    def add_vulnerability(self, vulnerability):
        """ Add a vulnerability to the scope. Automatically sort the vulnerabilities after adding. """
        self.vulnerabilities.append(vulnerability)
        if vulnerability.nrisk > self.highestrisk:
            self.highestrisk = vulnerability.nrisk
        if vulnerability.sectionweight > self.sectionweight:
            self.sectionweight = vulnerability.sectionweight

        self.vulnerabilities.sort()
    
    def set_introduction(self, introduction_headers, introduction):
        self.introduction_headers = introduction_headers
        self.introduction = introduction
        # Process the introduction section weight if it's there
        if "sectionweight" in self.introduction_headers and self.introduction_headers["sectionweight"] > self.sectionweight:
            log.info(f"Setting sectionweight for {self.name} to {self.introduction_headers['sectionweight']} based on introduction header")
            self.sectionweight = self.introduction_headers["sectionweight"]

    def get_list(self, validation=None):
        vulnlist = []
        for vuln in self.vulnerabilities:
            if validation == None:
                vulnlist.append(vuln.get_list())
            else:
                vl = vuln.get_list()
                validator = cerberus.Validator()
                validator.allow_unknown = True
                if vuln.validate(vl):
                    vulnlist.append(vuln.get_list())
        return vulnlist

    def get_ref(self, vulnerability_name):
        """ Get the reference of a vulnerability by name. Usually called from a markdown file. """
        for vulnerability in self.vulnerabilities:
            if vulnerability_name.lower() in vulnerability.name.lower():
                return vulnerability.ref

        return None

    def nvulns(self, riskrating=""):
        """ Get the number of vulnerabilities with the given risk rating """
        return sum(riskrating in vuln.risk for vuln in self.vulnerabilities)

    def nriskdict(self):
        """ Get a dictionary with risk:nrisks for all vulnerabilities """
        return Counter(vulnerability.risk for vulnerability in self.vulnerabilities)

    def updated(self, date=datetime.date.today()):
        """ Process the 'updates' header """
        newscope = copy.copy(self)

        # Regenerate with the added vulnerabilities
        newscope.vulnerabilities = []
        for vuln in self.vulnerabilities:
            newscope.add_vulnerability(vuln.updated(date))

        return newscope

    def generate_markdown(self, env):
        for vuln in self.vulnerabilities:
            vuln.generate_markdown(env)
        if self.introduction != "":
            self.introduction = process_template(self.introduction_headers, self.introduction, env)


class Vulnerability:
    """A class for vulnerability markdown.

    Has the following attributes:

     - name : The name of the vulnerability as a string
     - vulnerability_markdown : The markdown of the vulnerability as a tuple (headers, markdown)
     - scope : The name of the scope as a string. Used by VulnerabilityList for sorting a bucket of vulnerabilities into scopes,
                 taken from the headers, but this may be overwritten if it's in a scope bucket already.
     - weight : A number, used for sorting vulnerabilities. Taken from the headers.
     - sectionweight : A number, used for increasing the weight of a scope for sorting them. Set for particularly bad vulns. Taken from the headers.
     - risk : The risk of the vuln as a string.
     - nrisk : The risk of the vuln as a number. Used for sorting vulnerabilities.
     - headers : The headers of the markdown.
    """

    def __init__(self, name, riskassessment, headers=dict(), markdown='', scope='', weight=0, sectionweight=0, filename=''):

        self.name = name
        self.vulnerability_markdown = markdown
        self.scope = scope
        self.ref = 0  # ref will be set by VulnerabilityList
        self.filename = filename
        self.riskassessment = riskassessment

        if "riskassessment" in headers:
            if riskassessment.id not in headers['riskassessment']:
                raise Exception("Risk assessment {} not in the risk assessment for vulnerability {}".format(
                    riskassessment.name, name))
            headers.update(headers['riskassessment'][riskassessment.id])

        if "scope" in headers:
            self.scope = headers["scope"]
        else:
            self.scope = ''

        if "weight" in headers:
            self.weight = float(headers["weight"])
        else:
            self.weight = 0

        if "sectionweight" in headers:
            self.sectionweight = float(headers["sectionweight"])
        else:
            self.sectionweight = 0

        self.risk = riskassessment.get_risk(headers)
        self.nrisk = next((i for i, x in enumerate(
            riskassessment.risks) if x == self.risk), 0)
        headers['risk'] = self.risk

        if "status" not in headers:
            headers['status'] = "Open"

        self.status = headers['status']
        self.originalstatus = self.status
        self.originalrisk = self.risk

        self.headers = headers

    def __lt__(self, other):
        if other.nrisk == self.nrisk:
            return other.weight < self.weight
        else:
            return other.nrisk < self.nrisk

    def get_list(self):
        vulnlist = {
            'name': self.name,
            'ref': self.ref,
            'status': self.status,
            'risk': self.risk,
            'nrisk': self.nrisk,
            'originalstatus': self.originalstatus,
            'originalrisk': self.originalrisk,
            'filename': self.filename,
            'headers': self.headers
        }
        for stage in self.riskassessment.stages:
            if stage['name'] in self.headers:
                vulnlist[stage['name']] = self.headers[stage['name']]
        return vulnlist

    def updated(self, date=datetime.date.today()):
        # Get a new vulnerability
        newvuln = copy.copy(self)

        # Deep headers needs a deep copy otherwise the updates will change the original vuln
        newvuln.headers = copy.deepcopy(self.headers)

        if 'updates' in self.headers:
            for update in make_list(self.headers['updates']):
                if 'date' not in update or update['date'] <= date or date == "all":
                    for newvar in update.keys():
                        if newvar == 'date':
                            continue
                        newvuln.headers[newvar] = update[newvar]

        # We need to re-evaluate the risk just in case the likelihood or consequence changed
        newvuln.risk = newvuln.riskassessment.get_risk(newvuln.headers)
        newvuln.headers['risk'] = newvuln.risk
        newvuln.status = newvuln.headers['status']

        return newvuln

    def generate_markdown(self, env):
        """ Get the markdown of the vulnerability, processing the template with the current env.

        Usually called from a markdown file. """
        # First update the ref
        self.headers['ref'] = self.ref
        self.headers['scope'] = self.scope

        try:
            self.markdown = process_template(self.headers, self.vulnerability_markdown,
                                             env, name="vulnerability {}".format(self.name), filename=self.filename)
        except jinja2.exceptions.TemplateSyntaxError as error:
            log.error("Jinja2 error processing vulnerability {}: {} at lineno {} for file {}".format(
                self.name, error.message, error.lineno, filename=self.filename))
            log.error("Removing markdown for vulnerability {}".format(self.name))
            log_jinja2_error(self.vulnerability_markdown, error)
            self.markdown = ""
        except Exception as e:
            log.error(
                "Received exception when processing markdown for vulnerability {}: {}".format(self.name, e.args))
            log.error(traceback.format_exc())
            log.error("Removing markdown for vulnerability {}".format(self.name))
            self.markdown = ""



import cerberus
import logging

log = logging.getLogger(__name__)


def make_list(value):
    """ If it's a list, return value, otherwise return [value]
    """
    if type(value) is list:
        return value
    else:
        return [value]


def filter_rows(table, validation, return_fails=False):
    """Filter table rows based on the supplied validation
    """
    validator = cerberus.Validator()
    validator.allow_unknown = True
    validator.schema = validation
    new_table = []
    for row in table:
        if validator.validate(row):
            if not return_fails:
                new_table.append(row)
        else:
            if return_fails:
                new_table.append(row)
    return new_table


def sort_table(table, column, mapper=None, reverse=False):
    """Sort the table based off content in the chosen column.
    """
    if len(table) == 0:
        return table
    if mapper:
        return sorted(table, key=lambda d: mapper[d[column]] if d[column] in mapper else d[column], reverse=reverse)
    return sorted(table, key=lambda d: d[column], reverse=reverse)


def table_aggregate_value(table, column, aggtype, map_values=None):
    """Get aggregate value from the supplied table.

    aggtype can be 'min', 'max', 'mean', 'sum', or 'count'.
    map_values is a dict {value:mapping}. It will substitute value for mapping if found in the column.
    """
    val = None
    sum = 0
    count = 0
    for row in table:
        if type(row) is dict:
            if column in row:
                cv = row[column]
            else:
                log.warn(f"Column {column} not in rows for table_aggregate value. Choices are: {row}")
                continue
        else:
            if column >= 0 and column < len(row):
                cv = row[column]
            else:
                log.warn(f"Column {column} not in rows for table_aggregate value. Row length is {len(row)}")
                continue

        if map_values:
            if cv in map_values:
                cv = map_values[cv]

        if aggtype == 'min':
            if val == None or cv < val:
                val = cv
        elif aggtype == 'max':
            if val == None or cv > val:
                val = cv
        elif aggtype == 'count':
            count += 1
        elif aggtype == 'sum':
            if type(cv) is int or type(cv) is float:
                sum += cv
            else:
                log.warn(f"table_aggregate_value removing {cv} from sum as it is not an integer")
        elif aggtype == 'mean':
            sum += cv
            count += 1
        else:
            log.warn(
                "Invalid aggtype encountered for table_aggregate_value: {}".format(aggtype))
            return None

    if aggtype == 'min' or aggtype == 'max':
        return val
    if aggtype == 'count':
        return count
    if aggtype == 'sum':
        return sum
    if aggtype == 'mean':
        return sum / count

    return None


def _get_aggregate(cgbv, prevrows, groupby, dict_list, agg, map_values):
    nr = cgbv

    # Do the aggregation
    for column, aggtype in agg.items():
        if dict_list:
            nr["{}_{}".format(column, aggtype)
               ] = table_aggregate_value(prevrows, column, aggtype, map_values)
        else:
            nr.append(table_aggregate_value(
                prevrows, column, aggtype, map_values))
    return nr


def table_aggregate(table, groupby, agg, map_values=None):
    """Get aggregate information from a table with groupings specified in groupby.

    groupby should be either a single column or a list of columns. If the table is a list of dicts the columns should be
    column heading names. If it is a list of lists it should be the index number of the table.
    """

    # Go backwards sorting with groupby so that we have everything sorted
    if type(groupby) is str or type(groupby) is int:
        # If it's a string make it a list of one
        groupby = [groupby]

    st = table
    for gb in reversed(groupby):
        st = sort_table(st, gb)

    if len(st) == 0:
        return []

    if type(st[0]) is dict:
        dict_list = True
    else:
        dict_list = False

    # We now have everything sorted by the groupby columns. Now we need to get the aggregates
    if dict_list:
        cgbv = {}
        for gb in groupby:
            cgbv[gb] = st[0][gb]
    else:
        cgbv = []
        for gb in groupby:
            cgbv.append(st[0][gb])
    prevrows = []

    new_table = []

    for row in st:
        for i in groupby:
            if row[i] != cgbv[i]:
                # We have a new row
                if len(prevrows) > 0:
                    # We're not at the start of the table so we've hit the next block to aggregate
                    # Initialise the new row with what's in groupby

                    new_table.append(_get_aggregate(
                        cgbv, prevrows, groupby, dict_list, agg, map_values))
                # Initialise the new values
                if dict_list:
                    cgbv = {}
                    for gb in groupby:
                        cgbv[gb] = row[gb]
                else:
                    cgbv = []
                    for gb in groupby:
                        cgbv.append(row[gb])
                prevrows = []

        # Add the row to prevrows
        prevrows.append(row)

    new_table.append(_get_aggregate(
        cgbv, prevrows, groupby, dict_list, agg, map_values))

    return new_table


def table_to_dict(table, keyindex=None, valindex=None, keys=None, default=None):
    """ Convert tabular data into a dict of key:value pairs taking the keys and values from the keyindex and valindex columns.

    This is used for feeding tabular data into charts.
    keys provides a filter for keys. If default is set, then keys that don't appear in the table will appear in the returned dict
    with the default value.
    """
    d = {}

    if keys != None and default != None:
        for k in keys:
            d[k] = default

    if len(table) == 0:
        # We have an empty table
        if keys != None and default != None:
            for key in keys:
                d[key] = default
            return d
        else:
            return {}

    if keyindex == None:
        if len(table[0]) < 1:
            log.warning(
                "keyindex for table_to_dict is not set and the first row has less than 1 value so it cannot be automatically set")
            return {}
        keyindex = list(table[0].keys())[0]

    if valindex == None:
        if len(table[0]) < 2:
            log.warning(
                "valindex for table_to_dict is not set and the first row has less than 2 values so it cannot be automatically set")
            return {}
        valindex = list(table[0].keys())[1]

    for row in table:
        if keyindex not in row or valindex not in row:
            continue
        if keys:
            if row[keyindex] not in keys:
                continue
        d[row[keyindex]] = row[valindex]

    return d


def tables_outer_join(table1, table2, column1, column2=None):
    """ Match the data in column1 for table1 to data in column2 for table2 to perform an outerjoin on two tables
    Tables can be either lists of lists or lists of dicts. If it is a list of lists then the column should be an int, otherwise
    it should refer to the dictionary key of the column.
    If column2 is not specified, column1 will be used for table2 as well.
    """
    new_table = []
    if len(table1) == 0:
        return []
    if type(table1[0]) is dict:
        dict_list = True
    else:
        dict_list = False

    for r1 in table1:
        for r2 in table2:
            if column2 != None:
                if r1[column1] == r2[column2]:
                    if dict_list:
                        new_table.append(r1 | r2)
                    else:
                        new_table.append(r1 + r2)
            else:
                if r1[column1] == r2[column1]:
                    if dict_list:
                        new_table.append(r1 | r2)
                    else:
                        new_table.append(r1 + r2)

    return new_table


def table_separate_column(table, column, separator, columnmapper=None, separatormapper=None, separator_list=None, dict_list=False):
    """ Separate out the content of one column into separate columns for each unique value of the content in separator.
    This is useful for displaying compliance stuff. For instance, you have a list of Essential 8 controls in a spreadsheet
    and you want a list of controls for each maturity level. You can use this to show them in three columns.

    If the content of separator is integers the output will be in order of the integer.

    columnmapper is used to change the content of the column in the output. For instance changing "Passed"/"Failed" to "P"/"F".
    separatormapper can be used in the same way for the separator. This can be used to map two pieces of content to the same
    thing or to map content to integers to have them ordered in the right way.
    separator_list can manually force the columns by using this list.
    """
    if separator_list is not None:
        separators = separator_list
    else:
        separators = []
        # First we need to go through and get all the separators
        for row in table:
            if separatormapper is not None and row[separator] in separatormapper:
                if separatormapper[row[separator]] not in separators:
                    separators.append(separatormapper[row[separator]])
            elif row[separator] not in separators:
                separators.append(row[separator])

        # We need to get the integers out and put them at the start in order
        orderedsep = []
        otherseps = []
        for sep in separators:
            if type(sep) is int:
                orderedsep.append(sep)
            else:
                otherseps.append(sep)

        orderedsep = sorted(orderedsep)
        separators = orderedsep + otherseps

    # We now have a list of separators with ints ordered at the front
    # To get the columns we have a dict of lists.
    columns = {}
    for sep in separators:
        columns[sep] = []

    # Let's put the column contents into the above data structure
    for row in table:
        if separatormapper is not None and row[separator] in separatormapper:
            sep = separatormapper[row[separator]]
        else:
            sep = row[separator]

        if sep not in separators:
            continue

        if columnmapper is not None and row[column] in columnmapper:
            col = columnmapper[row[column]]
        else:
            col = row[column]

        columns[sep].append(col)

    # We now have the columns in order. Create a new table.
    new_table = []
    finished = False
    while not finished:
        finished = True
        if dict_list:
            new_row = {}
        else:
            new_row = []

        for sep in columns:
            if len(columns[sep]) == 0:
                nc = ""
            else:
                nc = columns[sep].pop(0)
                finished = False

            if dict_list:
                new_row[sep] = nc
            else:
                new_row.append(nc)

        if not finished:
            new_table.append(new_row)

    return new_table


def table_separate_column_groups(table, groupby, column, separator, columnmapper=None, separatormapper=None, separator_list=None, dict_list=False):
    """table_separate_column but it 

    groupby is used to separate out into groups. This way each group will start at the same level.
    """
    # Get a list of groups
    groups = {}
    for row in table:
        if groupby not in row:
            gb = None
        else:
            gb = row[groupby]
        if gb == "None":
            gb = 'default_table_separate_column_groups'
        if gb not in groups:
            groups[gb] = []
        groups[gb].append(row)

    new_table = []
    for gb, group in groups.items():
        new_table += table_separate_column(
            group, column, separator, columnmapper, separatormapper, separator_list, dict_list)
    return new_table


def table_add_row(table, row=None, index=None):
    """Add a row to the table. If row is None it adds an empty row. If index is None it adds it to the end.
    """
    if row == None:
        if len(table) == 0:
            return [[]]
        if type(table[0]) is dict:
            row = {}
        else:
            row = []

    if index > len(table):
        table.append(row)
    else:
        table.insert(index, row)
    return table


def separate_sequences(sequence):
    """Separates out a dict into repeated keys. Useful for coloured charts.

    For example:
      {'a':'1', 'b':'2'}
    becomes
      {'a':{'a':'1'}, {'b':{'b':'2'}}
    """

    newdict = {}
    for k, i in sequence.items():
        newdict[k] = {k: i}
    return newdict

def table_colpicker(table, colpicker):
    """Take only specific columns from a table. Column is either the first row or dictionary keys.
    """
    if len(table) == 0:
        return table
    
    if not isinstance(colpicker, list):
        log.warn("Colpicker variable not a list, ignoring.")
        return table
    
    
    # Check whether the table is a list of dicts or list of lists
    if isinstance(table[0], dict):
        newtable = []
        for i in range(len(table)):
            newtable.append({})
        for col in colpicker:
            # Is the column a string?
            if not isinstance(col, str):
                log.warn(f"Entry in colpicker is not a string. Ignoring. Found: {col}")
                continue
            if col not in table[0]:
                log.warn(f"Column {col} not a heading in the table. The headings are: {table[0].keys()}")
                continue
            for trow in range(len(table)):
                newtable[trow][col] = table[trow][col]
    else:
        newtable = []
        for i in range(len(table)):
            newtable.append([])
        for col in colpicker:
            # Is the column a string?
            if isinstance(col, str):
                if col in table[0]:
                    col = table[0].index(col)
                else:
                    log.warn(f"Entry {col} in colpicker is not a heading. Ignoring. Headings are: {table[0]}")
                    continue
            
            if not isinstance(col, int):
                log.warn(f"Entry {col} in colpicker is not an integer or string. Ignoring.")
                continue

            if col < 0 or col >= len(table[0]):
                log.warn("Entry in colpicker outside the range of the table length {}. Found: {}".format(
                    len(table[0]), col))
                continue

            for trow in range(len(table)):
                newtable[trow].append(table[trow][col])

    return newtable



import os
from pathlib import Path
import logging
import csv
import traceback
import copy

try:
    from openpyxl import load_workbook
    openpyxl_imported = True
except ImportError:
    openpyxl_imported = False

log = logging.getLogger(__name__)


def import_csv(csvfile, as_dict_list=False, index_col=None):
    filepath = Path(csvfile)
    cwdpath = Path(os.path.curdir)
    if cwdpath not in filepath.parents:
        log.warning(
            "import_csv figure path {} doesn't seem to be in cwd".format(csvfile))
        return ''

    try:
        with open(filepath) as csvfileopen:
            reader = csv.reader(csvfileopen)
            return make_structure(reader, as_dict_list, index_col)
    except Exception as e:
        log.warning(
            "Error opening CSV file {}: {}".format(filepath, e.args))
        return ''


def import_xlsx(fp, worksheet=None, min_row=None, max_row=None, min_col=None, max_col=None, as_dict_list=False, index_col=None):
    if worksheet:
        log.info(f"Importing XLSX document {fp} worksheet {worksheet}")
    else:
        log.info(f"Importing XLSX document {fp}")
    if not openpyxl_imported:
        log.warning(
            "Unsuccessful loading of openpyxl, skipping import of XLSX file")
        return []
    filepath = Path(fp)
    cwdpath = Path(os.path.curdir)
    if cwdpath not in filepath.parents:
        log.warning(
            "import_xlsx path {} doesn't seem to be in cwd".format(fp))
        return ''

    try:
        workbook = load_workbook(filename=fp)
        if worksheet:
            sheet = workbook[worksheet]
        else:
            sheet = workbook.active

        row_iterator = sheet.iter_rows(
            min_row, max_row, min_col, max_col, values_only=True)
        return make_structure(row_iterator, as_dict_list, index_col)
    except Exception as e:
        log.warning(
            "error importing XLSX file {}: {}".format(fp, e.args))
        log.warn(traceback.format_exc())

        return ''


def make_structure(row_iterator, as_dict_list=False, index_col=None):
    # Let's just make 0 and 1 the same thing for simplicity and less user errors
    if index_col == 0:
        index_col = 1

    if index_col:
        results = {}
    else:
        results = []

    headings = []
    for row in row_iterator:
        # Make sure row is a list and not a tuple
        row = copy.deepcopy(list(row))

        # We need to store the headings to use as keys in our dicts.
        if as_dict_list and len(headings) < 1:
            headings = row
            continue

        # Turn this row into a dict, rather than a list.
        if as_dict_list:
            new_item = {}
            i = 0
            for key in headings:
                new_item[key] = row[i]
                i += 1
        else:
            new_item = row

        # We either create a new key/value pair or add the item to the list
        if index_col:
            results[row[index_col-1]] = new_item
        else:
            results.append(new_item)

    return results



import yaml
import cerberus
import logging
import os

log = logging.getLogger(__name__)

# Cerberus validation for a strict dictionary
validate_strictdict = {
         'propertyschema': {'type': 'string'},
         'valuesschema':{
            'type':'string'
         }
     }

def process_templatemapperfile(filename):
    mapper = {}
    directory = os.path.dirname(filename)
    try:
        with open(filename, "r") as stream:
            try:
                filemapper = yaml.safe_load(stream)

                # Validate that the mapper is a simple dictionary
                v = cerberus.Validator(validate_strictdict)
                v.validate(filemapper)
                
                for name, location in filemapper.items():
                    mapper[name] = os.path.join(directory, location)
                
            except yaml.YAMLError as error:
                log.warn(f"Template mapper file '{filename}' not valid YAML code, gives error {error.args}")
                return {}
            except cerberus.schema.SchemaError as error:
                log.warn(f"Template mapper file '{filename}' does not contain a simple dict. Validation error {error.args}. Skipping.")
                return {}
    except FileNotFoundError as error:
        log.warn(f"Template mapper file '{filename}' not found. Skipping.")
    
    return mapper
    


def process_templatemapper(arg = None, config = {}, configfiles = []):
    """ Process the template mappers provided (if any).

    The order of precedence will be arg -> config -> configfiles.
    arg: a string containing the location of a template mapper or None
    config: A dictionary mapper from the config.py file
    configfiles: A list of files containing locations of template mappers
    """
    mapper = {}
    if configfiles:
        for tm in configfiles:
            mapper.update(process_templatemapperfile(tm))
    
    if config:
        mapper.update(config)

    if arg != None and arg != '':
        mapper.update(process_templatemapperfile(arg))

    return mapper


import os
from report_ranger.utils.mdread import markdown_from_file
import logging

log = logging.getLogger(__name__)

def retrieve_template(templatefile, templatemapper={}, parentdir=""):
    if templatefile in templatemapper:
        return Template(templatemapper[templatefile], templatemapper)
    if parentdir:
        templatefile = os.path.join(parentdir, templatefile)
    if os.path.isfile(templatefile):
        return Template(templatefile, templatemapper)
    raise Exception(f"Could not find template {templatefile}")

class Template:
    template_headers={}
    template_markdown=""

    def __init__(self, templatefile, templatemapper={}):
        self.templateheaders, self.templatemarkdown = markdown_from_file(
            templatefile,
            includemapper=templatemapper,
            pathlist=[os.path.dirname(templatefile)]
            )

        self.templatedir = os.path.dirname(os.path.abspath(os.path.join(os.path.curdir, templatefile)))
        self.templateheaders['templatedir'] = self.templatedir

        self.templateheaders['latex_template'] = os.path.join(
            self.templatedir, self.templateheaders['latex']['template'])

        if 'html' in self.templateheaders:
            self.templateheaders['html_template'] = os.path.join(
                self.templatedir, self.templateheaders['html']['template'])
        else:
            self.html_template = ''



import os
import jinja2
from report_ranger.utils.jinja_helpers import log_jinja2_error
from report_ranger.utils.mdread import markdown_from_file, process_template
from report_ranger.table import Table, convert_list_of_dicts_to_matrix
from report_ranger.contentassistant import ContentAssistant
from num2words import num2words
import pandas as pd
import plotly.express as px
from pathlib import Path
from report_ranger.environment import Environment
from tabulate import tabulate
from report_ranger.imports.imports import import_csv, import_xlsx
import traceback
import logging
import ipaddress
import re
from report_ranger.styles import TableStyles, ColorList, test_color
from report_ranger.imports.section import Section
import plotly.io as pio
import plotly.graph_objects as go
from report_ranger.helpers import table_aggregate, table_aggregate_value, filter_rows, table_to_dict, tables_outer_join, table_separate_column, table_add_row, table_separate_column_groups, sort_table, separate_sequences, table_colpicker

log = logging.getLogger(__name__)
if pio:
    pio.kaleido.scope.mathjax = None

headeralias = {
    'h': 'thead',
    'e': 'temph',
    'L': 'tlow',
    'M': 'tmedium',
    'H': 'thigh',
    'C': 'tcritical',
    'I': 'tinformational',
    'o': 'topen',
    'c': 'tclosed',
    'cb': 'tclosedbold',
    'O': 'topenbold',
    'ob': 'topenbold',
    'b': 'bold'
}

figurecache = {}

def oxfordcomma(listed):
    if len(listed) == 0:
        return ''
    if len(listed) == 1:
        return listed[0]
    if len(listed) == 2:
        return listed[0] + ' and ' + listed[1]
    return ', '.join(listed[:-1]) + ', and ' + listed[-1]

def validate_fqdn(fqdn):
        pattern = re.compile("^((?!-)[A-Za-z0-9-]{1,63}(?<!-)\\.)+[A-Za-z]{2,6}$")
        return pattern.match(fqdn) is not None

class OutputFormatter:
    def __init__(self, templateheaders=dict(), timer=None, watcher=None):
        self.templateheaders = templateheaders
        self.figformat = "png"
        self.watcher=watcher
        log.debug(f"watcher is {watcher}")

        self.timer = timer
        self.watcher=watcher
        log.debug(f"watcher is {watcher}")

        self.timer = timer

        self.timer = timer

        # Initial filling in the environment
        log.debug(f"{self.time()}: Initialising new environment for output formatter")
        env = Environment()
        env.set_static('file_exists', self.file_exists)
        env.set_static('include_file', self.include_file)
        env.set_static('raw_output', self.raw_output)
        env.set_static('escape', self.escape)
        env.set_static('stringescape', self.escape)
        env.set_static('pathjoin', os.path.join)
        env.set_static('range', range)
        env.set_static('str', str)
        env.set_static('int', int)
        env.set_static('len', len)
        env.set_static('num2words', num2words)
        env.set_static('oxfordcomma', oxfordcomma)
        env.set_static('px', px)
        env.set_static('pd', pd)

        env.set_static('nfig', 0)
        ca = ContentAssistant()
        env.set_static('ca', ca)
        env.set_static('ft', ca.format_table)
        env.set_static('format_table', ca.format_table)
        env.set_static('counter', ca.counter)
        env.set_static('get_counter', ca.get_counter)
        env.set_static('reset_counter', ca.reset_counter)
        env.set_static('icounter', ca.icounter)
        env.set_static('get_icounter', ca.get_icounter)
        env.set_static('reset_icounter', ca.reset_icounter)
        env.set_static('table_row', ca.table_row)
        env.set_static('table_rows', ca.table_rows)
        env.set_static('display_table', ca.display_table)

        env.set_static('newpage', self.newpage)
        env.set_static('new_page', self.newpage)
        env.set_static('import_csv', import_csv)
        env.set_static('csv', import_csv)
        env.set_static('import_xlsx', import_xlsx)
        env.set_static('xlsx', import_xlsx)
        env.set_static('import_section', self.import_section)
        env.set_static('process_tags', self.process_tags)
        env.set_static('build_from_data', self.build_from_data)
        env.set_static('new_section', self.newsection)
        env.set_static('end_report', self.end)
        env.set_static('iplist', self.iplist)
        env.set_static('table', self.table)
        env.set_static('figure', self.fig)
        env.set_static('fig', self.fig)
        env.set_static('bar_chart', self.bar_chart)
        env.set_static('polar_bar_chart', self.polar_bar_chart)
        env.set_static('funnel_chart', self.funnel_chart)
        env.set_static('chart', self.chart)
        env.set_static('include', self.include_file)
        env.set_static('table_aggregate', table_aggregate)
        env.set_static('table_aggregate_value', table_aggregate_value)
        env.set_static('filter_rows', filter_rows)
        env.set_static('sort_table', sort_table)
        env.set_static('table_to_dict', table_to_dict)
        env.set_static('tables_outer_join', tables_outer_join)
        env.set_static('table_separate_column', table_separate_column)
        env.set_static('table_add_row', table_add_row)
        env.set_static('table_separate_column_groups',
                       table_separate_column_groups)
        env.set_static('separate_sequences',
                       separate_sequences)
        env.set_static('table_colpicker', table_colpicker)
        env.set_static('convert_list_of_dicts_to_matrix', convert_list_of_dicts_to_matrix)

        # Legacy variables - these function names have changed, but can be referred to with the old function name
        env.set_static('include_file', self.include_file)  # Legacy

        # For legacy reasons, let's set a new "of" variable which contains pointers to functions in the output formatter
        of = type('of', (object,), {})()
        of.newpage = self.newpage
        of.newsection = self.newsection
        of.end = self.end
        of.table = self.table
        of.iplist = self.iplist
        of.fig = self.fig
        of.include_file = self.include_file
        of.escape = self.escape
        env.set_static('of', of)
        
        log.debug(f"{self.time()}: Finished basic setting up env")

        # Start with the defaults
        if 'defaults' in templateheaders:
            env.set_variables(templateheaders['defaults'])

        # Get color list
        self.colors = ColorList()
        if 'colors' in self.templateheaders:
            self.colors.addcolors(self.templateheaders['colors'])
        if 'colors' in self.templateheaders['riskassessment']:
            self.colors.addcolors(self.templateheaders['riskassessment']['colors'])


        log.debug(f"{self.time()}: Setting up table styles")

        # Get table styles
        self.tablestyles = TableStyles()

        temp_ts = dict()

        if 'table_styles' in self.templateheaders:
            temp_ts = self.templateheaders['table_styles']

        if 'table_styles' in self.templateheaders['riskassessment']:
            temp_ts.update(self.templateheaders['riskassessment']['table_styles'])

        for name, style in temp_ts.items():
            # We need to make sure the colors are in the colorlist otherwise
            # latex won't be able to refer to the color
            if 'color' in style and not self.colors.getcolor(style['color']):
                style['color'] = self.colors.addcolor(None, style['color'])
                log.debug("added new color {}".format(style['color']))
            if 'bgcolor' in style and not self.colors.getcolor(style['bgcolor']):
                style['bgcolor'] = self.colors.addcolor(
                    None, style['bgcolor'])
                log.debug("added new color {}".format(style['bgcolor']))

            self.tablestyles.addstyle(name, style)

        self.env = env
        
        log.debug(f"{self.time()}: Setting up plotly template")

        self.setup_plotly_template(self.templateheaders, self.colors)

        log.debug(f"{self.time()}: Finished output formatter initialisation")

    def raw_output(self, output):
        ca = self.env.get('ca')
        return ca._register_string(output)

    def time(self):
        if self.timer:
            return self.timer.time()
        return ""

    def stringescape(self, text):
        return text.replace("\\", "\\\\")

    def escape(self, text):
        return text

    def import_section(self, directory, ordinal='1'):
        newsection = Section(directory, ordinal, watcher=self.watcher)
        return newsection.get_chapters(self.env)

    def iplist(self, listofips, heading="IP list", columns=4):
        '''Build a table from a list of IPs. Automatically format based on number of columns and sort IPs into order.'''

        log.debug("iplist(): Building IP list")

        # Separate IPv4 and IPv6 addresses
        ipv4_nets = []
        ipv6_nets = []
        fqdns = []

        for _ip in listofips:
            try:
                if validate_fqdn(_ip):
                    fqdns.append(_ip)
                else:
                    ip_net = ipaddress.ip_network(_ip)
                    if ip_net.version == 4:
                        ipv4_nets.append(ip_net)
                    elif ip_net.version == 6:
                        ipv6_nets.append(ip_net)
            except Exception as e:
                    print("Invalid IP address or subnet or FQDN: {}. Error: {}".format(_ip,e))

        # Collapse and sort IPv4 and IPv6 addresses separately
        sorted_ipv4_nets = list(ipaddress.collapse_addresses(ipv4_nets))
        sorted_ipv6_nets = list(ipaddress.collapse_addresses(ipv6_nets))
        sorted_fqdns = sorted(fqdns)

        # Combine sorted lists
        sorted_nets = sorted_fqdns + sorted_ipv4_nets + sorted_ipv6_nets

        sortedips = [str(net) for net in sorted_nets]

        if(len(sortedips) < columns):
            columns = len(sortedips)

        iptable = [[heading]]
        for i in range(len(sortedips)):
            row = int(i / columns) + 1
            if i % columns == 0:
                iptable += [[]]
            iptable[row] += [sortedips[i]]

        return self.table(iptable, headings='top', colspan=[[columns]], cellalign=[['c']])

    def table(self, table, options={}, **tableargs):
        ''' This function formats the table in either latex or markdown, depending on the output.'''

        t = Table(table, env=self.env, **tableargs)

        # We're going to do a markdown table with tabulate
        if t.colalign != None:
            # We need to convert the 'l' and 'r' in colalign if it exists
            ca = []
            for col in t.colalign:
                if col == 'r':
                    ca.append('right')
                elif col == 'c':
                    ca.append('center')
                else:
                    ca.append('left')
        else:
            ca = None

        return tabulate(t.table[1:], t.table[0], stralign=ca, tablefmt="github")

    def newsection(self):
        return ""

    def newpage(self):
        return ""

    def headers(self):
        return ""

    def end(self):
        """ To be called at the end of the output - puts in the ending page/graphics """
        return ""

    def ra_methodology(self, ra):
        """ The output for the risk assessment methodology for the appendices """
        markdown = ra.markdown
        return markdown

    def file_exists(self, file_loc):
        fn, ext = os.path.splitext(file_loc)
        if ext != ".md" and ext != ".rr":
            return False
        return os.path.isfile(file_loc)

    def process_tags(self, text, add_to_env={}):
        if type(text) is not str:
            log.warn("Text passed to process_tags is not a string")
            return ""
        if text == "":
            log.warn("Empty string passed to process_tags")
            return ""
        return process_template(
            add_to_env, text, env=self.env, name="Call to process")

    def build_from_data(self, data, headings={}, heading_text={}, text={}, pre_text={}):
        """
        Build a hierarchy of report headings and text based on the entries in data.

        This function is commonly used to "reportise" a spreadsheet and so the expected data structures match
        the csv and xslx functions. Data should be in the format of dicts/lists of dicts/lists, like what the
        csv and xslx functions return.

        The headings parameter should provide a dict mapping between the depth of heading (1-4) to either the
        index or key of the column which should map to the heading. The text of the heading is pasted in, which
        means multiple lines will cause normal markdown to be displayed. Similarly, text is a dict mapping the
        heading depth with what text to show under that heading.
        """

        # If the structure is a dict, assume it's a row indexed dict. In that case we only want the values
        # as a list
        if type(data) is dict:
            data = data.values()

        # Make sure we now have a list
        if type(data) is not list:
            log.warn(
                "Data passed to build_from_data isn't a list or a dict, skipping.")
            return ''

        # Detect if we've got a list of dicts (colindexed) or a list of lists (not colindexed)
        if len(data) == 0:
            log.warn("Empty set of data passed to build_from_data, skipping.")
            return ''
        elif type(data[0]) is dict:
            colindexed = True
        elif type(data[0]) is list:
            colindexed = False
        else:
            log.warn(
                "Data passed to build_from_data is a list of something other than dicts or lists, skipping.")
            return ''

        try:
            # If we don't have accurate heading keys, the loop below will fail miserably. We've got to fix that.
            depths = list(headings.keys())
            for depth in depths:
                if depth not in range(1, 5):
                    log.warn(
                        "Heading keys in call build_from_data included {} which is not in the range 1-4, skipping".format(depth))
                    del headings[depth]

            # Lets try this again
            depths = list(headings.keys())

            if len(headings) == 0:
                log.warn("Call to build_from_data has no headings, skipping.")
                return ''

            current_headings = []
            markdown = ''
            for cr in range(len(data)):
                row = data[cr]

                for cd in range(len(depths)):
                    # Get the value of the heading. Note that this is the same whether it's a dict or list
                    heading = headings[depths[cd]]

                    if cd >= len(current_headings):  # We have to add the new heading
                        current_headings.append(row[heading])
                    elif row[heading] != current_headings[cd] and row[heading] != '' and row[heading] != None:
                        # We have a new heading. Let's snip off the headings of lower depths
                        # and add the new current heading
                        current_headings = current_headings[:cd] + [
                            row[heading]]
                    else:
                        # This heading isn't new, we can skip this depth
                        continue

                    # Need to pull pull together this and what is below
                    rd = dict(row)
                    th = {}
                    for key in rd:
                        new_key = str(key)
                        th[new_key.replace(
                            ' ', '_')] = rd[new_key]

                    # If there's pre_text print it out
                    # TODO: Add the potential to call other cells etc
                    if depths[cd] in pre_text:
                        markdown += "{}\n\n".format(
                            process_template(
                                th, pre_text[depths[cd]], env=self.env, name="build_from_data cell")
                        )

                    if depths[cd] in heading_text:
                        markdown += "{}\n\n".format(
                            process_template(
                                th, heading_text[depths[cd]], env=self.env, name="build_from_data cell")
                        )
                    else:
                        # Print it out
                        markdown += "{} {}\n\n".format("#" *
                                                       depths[cd], row[heading])

                    # If there's text print this out as well
                    if depths[cd] in text:
                        try:
                            # We need to get the rows under this heading to add to the jinja environment. That allows us
                            # to refer to the rows for summary lists for instance
                            # TODO This isn't necessary when you're referring directly to a column. Fix.
                            upcoming_rows = []
                            for ucr in range(cr, len(data)):
                                upcoming_row = data[ucr]
                                new_heading = False
                                for ucd in range(0, cd+1):
                                    heading = headings[depths[ucd]]
                                    if upcoming_row[heading] != current_headings[ucd] and upcoming_row[heading] != '' and upcoming_row[heading] != None:
                                        # log.info(
                                        #    "upcoming rows break: current heading {} is {}, upcoming_row[heading] is {}".format(ucd, current_headings[ucd], upcoming_row[heading]))
                                        new_heading = True
                                        break

                                if new_heading:
                                    break
                                else:
                                    upcoming_rows.append(upcoming_row)

                            if colindexed:
                                row_dict = dict(row)
                                # Fix up headings with spaces
                                text_headers = {}
                                for key in row_dict:
                                    new_key = str(key)
                                    text_headers[new_key.replace(
                                        ' ', '_')] = row_dict[new_key]
                                text_headers['upcoming_rows'] = upcoming_rows

                                # The user has the choice of putting in the title of a row or directly putting in templated markdown into text.
                                if text[depths[cd]] in row and row[text[depths[cd]]] != '':
                                    # Skip if there's nothing to display. This stops, for instance, displaying "None".
                                    if row[text[depths[cd]]] != '' and row[text[depths[cd]]] != None:
                                        markdown += "{}\n\n".format(process_template(
                                            text_headers, row[text[depths[cd]]], env=self.env, name="build_from_data cell"))
                                else:
                                    markdown += "{}\n\n".format(process_template(
                                        text_headers, text[depths[cd]], env=self.env, name="build_from_data cell"))

                                # log.info('upcoming_rows: {}'.format(upcoming_rows))

                            else:
                                text_headers = {'row': row,
                                                'upcoming_rows': upcoming_rows}

                                # Here the user can instead put in a number referring to the column number
                                if text[depths[cd]] in range(len(row)):
                                    # Skip if there's nothing to display. This stops, for instance, displaying "None".
                                    if row[text[depths[cd]]] != '' and row[text[depths[cd]]] != None:
                                        markdown += "{}\n\n".format(process_template(
                                            text_headers, row[text[depths[cd]]], env=self.env, name="build_from_data cell"))
                                else:
                                    markdown += "{}\n\n".format(process_template(
                                        text_headers, text[depths[cd]], env=self.env, name="build_from_data cell"))
                        except Exception as e:
                            log.warn(
                                "Exception occurred in building text in column {} of data in build_from_data, args: {}".format(heading, e.args))
                            log.warn(traceback.format_exc())

            return markdown

        except Exception as e:
            log.warn("Call to build_from_data failed: {}".format(e.args))
            log.warn(traceback.format_exc())
            return ''

    def include_file(self, file_loc, include_headers=True, custom_headers=None):
        """ Process the file, template it, and return just the markdown """
        fn, ext = os.path.splitext(file_loc)
        if ext != ".md" and ext != ".rr":
            return ""
        try:
            headers, markdown = markdown_from_file(file_loc, env=self.env, watcher=self.watcher)
            headers['cwd'] = os.path.dirname(
                os.path.join(os.path.curdir, file_loc))
        except:
            log.warn(f"Could not read included file {file_loc}")
            return ""

        if not include_headers:
            headers = {}

        if custom_headers != None:
            headers.update(custom_headers)

        try:
            output = process_template(
                headers, markdown, env=self.env, name="included file", filename=file_loc)
            return output
        except jinja2.exceptions.TemplateSyntaxError as error:
            log.error("Jinja2 error processing included file {}: {} at lineno {} for file {}".format(
                file_loc, error.message, error.lineno, filename=error.filename))
            log.error("Removing markdown for included file {}".format(file_loc))
            log_jinja2_error(markdown, error)
            log.error(traceback.format_exc())
            return ""
        except Exception as e:
            log.error(
                "Received exception when processing markdown for included file {}: {}".format(file_loc, e.args))
            log.error(
                "Removing markdown for included file {}".format(file_loc))
            log.error(traceback.format_exc())
            return ""

    def setup_plotly_template(self, template, colors):
        """ Setup the plotly template based on the table_styles template header
        """
        if 'charts' in template:
            log.debug(f"{self.time()}: Customising chart template")
            charts = template['charts']
            # Translate colours for the template
            for l in ['colorway']:
                if l in charts:
                    for i in range(len(charts[l])):
                        if colors.getcolor(charts[l][i]):
                            charts[l][i] = '#' + colors.getcolor(charts[l][i])
                        elif test_color(charts[l][i]):
                            charts[l][i] = '#' + charts[l][i]


            plotly_template = dict(
                layout=go.Layout(template['charts'])
            )
            pio.templates["reportranger"] = plotly_template

            log.debug(f"{self.time()}: Before set static plotly template")
            pio.templates.default = 'plotly_white+reportranger'
        else:
            plotly_template = 'plotly_white'
            pio.templates.default = 'plotly_white'

        self.env.set_static("plotly_template", plotly_template)

    def _dataframe_from_data(self, data, labels=['number', 'stage', 'category']):
        frames = []
        # Concatenate the data frames
        for i in data:
            df = pd.DataFrame(
                {
                    labels[0]: data[i].values(),
                    labels[1]: data[i].keys()
                }
            )
            df[labels[2]] = i
            frames.append(df)

        dataframe = pd.concat(frames, axis=0)
        return dataframe
    
    def _same_as_figurecache(self, nfig, values):
        log.debug("Figuring out figurecache")
        if nfig in figurecache:
            samefig = True
            for value in values:
                if value not in figurecache[nfig] or figurecache[nfig][value] != values[value]:
                    samefig = False
                    log.debug(f"Value {value} is not the same: {figurecache[nfig].get(value)} vs {values[value]}")
                    break
            
            if samefig:
                return True
            else:
                figurecache[nfig] = values
                return False
        else:
            log.debug(f"Registering figurecache {nfig}")
            figurecache[nfig] = values
            return False

    def bar_chart(self, description, data, update_layout=None, update_traces=None, *args, **kwargs):
        """ Simple vertical bar chart that is {column: value}
        """
        nfig = self.env.get('nfig')
        filename = 'screenshots/fig{}.{}'.format(int(nfig), self.figformat)
        if self._same_as_figurecache(nfig, {"description":description, "data":data, "update_layout":update_layout, "update_traces":update_traces, "args":args, **kwargs}):
            self.env.set_static('nfig', nfig + 1)
            return '![{}]({})'.format(description, filename)

        multi = False
        # See if we have multiple categories
        for i in data:
            if type(data[i]) is dict:
                multi = True
                break

        if multi:
            dataframe = self._dataframe_from_data(
                data, ['value', 'index', 'category'])
            figure = px.bar(dataframe, x="index", y="value",
                            color="category", *args, **kwargs)
        else:
            dataframe = pd.Series(data=data)

            figure = px.bar(dataframe, *args, **kwargs)
            figure.update_layout(showlegend=False)
        return self.fig(description, figure, update_layout, update_traces)

    def funnel_chart(self, description, data, *args, **kwargs):
        nfig = self.env.get('nfig')
        filename = 'screenshots/fig{}.{}'.format(int(nfig), self.figformat)
        if self._same_as_figurecache(nfig, {"description":description, "data":data, "args":args, **kwargs}):
            self.env.set_static('nfig', nfig + 1)
            return '![{}]({})'.format(description, filename)

        multi = False
        # See if we have multiple categories
        for i in data:
            if type(data[i]) is dict:
                multi = True
                break

        if multi:
            dataframe = self._dataframe_from_data(data)
            figure = px.funnel(dataframe, x="number", y="stage",
                               color="category", *args, **kwargs)
        else:
            dataframe = pd.DataFrame(
                {
                    'number': data.values(),
                    'stage': data.keys()
                }
            )

            figure = px.funnel(dataframe, x="number",
                               y="stage", *args, **kwargs)

        return self.fig(description, figure)

    def polar_bar_chart(self, description, data, colors=None, range=None, *args, **kwargs):
        """ Polar chart. data is a dict {category: value}, colors is a map {value:color}

        This function exists cause I really like the stacked colours for compliance work.
        """
        nfig = self.env.get('nfig')
        filename = 'screenshots/fig{}.{}'.format(int(nfig), self.figformat)
        if self._same_as_figurecache(nfig, {"description":description, "data":data, "colors":colors, "range":range, "args":args, **kwargs}):
            self.env.set_static('nfig', nfig + 1)
            return '![{}]({})'.format(description, filename)
        direction = []
        value = []
        maturity = []

        for category, level in data.items():
            level = int(level)
            if level <= 0:
                direction.append(category)
                value.append(0)
                maturity.append('1')
            else:
                if level > 100:
                    log.warn(
                        "Level {} passed to polar_bar_chart > 100".format(level))
                else:
                    cl = 1
                    while level > 0:
                        direction.append(category)
                        if level >= 1:
                            value.append(1)
                        else:
                            value.append(level)
                        maturity.append(str(cl))
                        cl += 1
                        level -= 1

        dataframe = pd.DataFrame(data={
            'direction': direction, 'value': value, 'maturity': maturity
        })

        if 'colors':
            kwargs['color_discrete_map'] = colors
            kwargs['color'] = 'maturity'

        figure = px.bar_polar(dataframe, r='value',
                              theta='direction', *args, **kwargs)

        update_layout = {'showlegend': False, 'polar_radialaxis_dtick': 1,
                         'polar_radialaxis_showticklabels': False,
                         'polar_radialaxis_showline': False
                         }

        if range:
            update_layout['polar_radialaxis_range'] = range

        figure.update_layout(**update_layout)

        return self.fig(description, figure)

    def chart(self, chart, description, data, update_layout=None, update_traces=None, *args, **kwargs):
        """ Generalist chart function
        """
        nfig = self.env.get('nfig')
        filename = 'screenshots/fig{}.{}'.format(int(nfig), self.figformat)
        if self._same_as_figurecache(nfig, {"description":description, "data":data, "args":args, "kwargs":kwargs}):
            self.env.set_static('nfig', nfig + 1)
            return '![{}]({})'.format(description, filename)
        
        multi = False
        # See if we have multiple categories
        for i in data:
            if type(data[i]) is dict:
                multi = True
                break

        if multi:
            dataframe = self._dataframe_from_data(
                data, ['value', 'index', 'category'])
            figure = chart(dataframe, *args, **kwargs)
        else:
            df = pd.DataFrame(
                {
                    'index': data.keys(),
                    'value': data.values()
                }
            )
            figure = chart(df, *args, **kwargs)
        return self.fig(description, figure, update_layout, update_traces)

    def fig(self, description, figure, update_layout=None, update_traces=None):
        nfig = self.env.get('nfig')
        self.env.set_static('nfig', nfig + 1)
        filename = 'screenshots/fig{}.{}'.format(int(nfig), self.figformat)
        log.info("Generating figure {}: {} to file {}".format(
            int(nfig), description, filename))
        filepath = Path(filename)
        cwdpath = Path(os.path.curdir)
        if cwdpath not in filepath.parents:
            log.warn(
                "Screenshot figure path {} doesn't seem to be in cwd".format(filepath))
            return ''

        if update_layout:
            figure.update_layout(**update_layout)
        if update_traces:
            figure.update_traces(**update_traces)

        figure.write_image(filename)
        return '![{}]({})'.format(description, filename)

    def _build_markdown_headers(self, headers):
        # Headers. This is a separate function because it might be modified or taken away by specific formatters.
        markdown = '''---\n'''
        markdown += headers
        markdown += "\n...\n\n"
        return markdown

    def _warn_in_text(self, templateheaders, env, text):
        ''' Throw a log warning if the warn list in templateheaders or env is in the text
        '''
        warn = []

        # Generate the warning list. This combines the lists in the template as well as the report (if it's there)

        # Get the list from the template
        if 'warn' in templateheaders:
            log.info("Warning on {}".format(templateheaders['warn']))

            if isinstance(templateheaders['warn'], list):
                warn = templateheaders['warn']
            else:
                log.warn(
                    "Warn header in the template front matter is not a list. Warning function won't work.")

        # Get the list from the report headers
        if 'warn' in env:
            if isinstance(env['warn'], list):
                warn = [*warn, *env['warn']]
            else:
                log.warn(
                    "Warn header in the report front matter is not a list. Warning function won't work.")

        # Go through the final output and warn on the line if the warning text is found
        for warntext in warn:
            if text.find(warntext):
                matched_lines = [line for line in text.split(
                    '\n') if warntext.lower() in line.lower()]
                for line in matched_lines:
                    log.warn(
                        "Found '{}' in the following line: {}".format(warntext, line))


    def _build_markdown(self, templatemarkdown):
        # A giant string to put all the output markdown into
        markdown = ""

        log.debug(f"{self.time()}: Outputting markdown")
        markdown += self._build_markdown_headers(self.headers())

        env = self.env.get_env()


        try:
            j2template = jinja2.Template(templatemarkdown)
            processedtemplate = j2template.render(env)
            markdown += processedtemplate

            # Content assistant parsing
            output = self.env.get('ca').parse_register(self, markdown)

            self._warn_in_text(self.templateheaders, env, output)


        except jinja2.exceptions.TemplateSyntaxError as error:
            log.error(
                "Error in processing the final template: {}".format(error.message))
            log_jinja2_error(markdown, error)
            raise Exception("Error reading the final template: {} at lineno {} for file {}".format(
                error.message, error.lineno, error.filename))
        except jinja2.exceptions.TemplateError as error:
            log.error(
                "Error in processing the final template: {}".format(error.message))
            raise Exception("Error reading the final template: {}".format(
                error.message))

        return output

    def output(self, markdown, outputfile=''):
        output = self._build_markdown(markdown)

        log.debug("Writing Markdown")

        with open(outputfile, 'w') as fh:
            fh.write(output)
        log.info("Finished writing")

        return output



import mistune
import re
import logging

log = logging.getLogger(__name__)


class LatexTableRenderer(mistune.HTMLRenderer):
    NAME = 'latex'
    IS_TREE = False

    def _escape_latex(self, text):
        ''' Escape the given text for display in Latex output

        :param text: a plain text message
        :return: the message escaped to appear correctly in LaTeX
        '''
        conv = {
            '&': r'\&',
            '%': r'\%',
            '$': r'\$',
            '#': r'\#',
            '_': r'\_',
            '{': r'\{',
            '}': r'\}',
            '~': r'{\textasciitilde}',
            '^': r'\^',
            '\\': r'{\textbackslash}',
            '<': r'{\textless}',
            '>': r'{\textgreater}',
        }
        regex = re.compile('|'.join(re.escape(str(key)) for key in sorted(
            conv.keys(), key=lambda item: - len(item))))
        return regex.sub(lambda match: conv[match.group()], text)

    def _escape_latex_codespan(self, text):
        ''' Escape the given text for codespans in latex. This has different rules to elsewhere.
        '''
        sepchar = None
        for char in "!\"'()*,-./:;_":
            if char not in text:
                sepchar = char
                break
        if not sepchar:
            return None
        conv = {
            '&': r'\&',
            '%': r'\%',
            '#': r'\#',
            '_': r'\_',
            '{': r'\{',
            '}': r'\}',
            '~': r'\~',
            '\\': r'\\\\'
        }
        regex = re.compile('|'.join(re.escape(str(key)) for key in sorted(
            conv.keys(), key=lambda item: - len(item))))
        return sepchar + regex.sub(lambda match: conv[match.group()], text) + sepchar

    def __init__(self, escape=True):
        super(LatexTableRenderer, self).__init__()
        self._escape = escape

    def text(self, text):
        return self._escape_latex(text)

    def link(self, text, url, title=None):
        if text is None or text == '':
            return '\\url{{{}}}'.format(self._escape_latex(url))

        s = '\\href{' + \
            self._escape_latex(url) + '}{' + self._escape_latex(text) + '}'
        return s

    def image(self, alt, url, title=None):
        s = '\\begin{{figure}}'
        s += '\\centering'
        s += '\\includegraphics{{{}}}}'.format(self._escape_latex(url))
        s += '\\caption{{{}}}'.format(self._escape_latex(alt))
        s += '\\end{{figure}}'
        return s

    def emphasis(self, text):
        return '\\emph{' + text + '}'

    def strong(self, text):
        return '\\textbf{' + text + '}'

    def codespan(self, text):
        code = '\\passthrough{\\lstinline' + \
            self._escape_latex_codespan(text) + '}'
        if code != None:
            return code
        return self.block_code(text)

    def linebreak(self):
        return ' \\\\ '

    def inline_html(self, html):
        return self._escape_latex(html)

    def paragraph(self, text):
        return text + ' \\\\ '

    def heading(self, text, level, **attrs):
        # TODO DOES NOT PROPERLY DO HEADINGS
        return text

    def newline(self):
        return ''

    def thematic_break(self):
        # TODO NO THEMATIC BREAK
        return ''

    def block_text(self, text):
        return text

    def block_code(self, code, info=None):
        code = code.replace('\n', '^^J\n')
        code = code.replace('{', '\{')
        code = code.replace('}', '\}')
        return f"""\\begin{{lstlisting}}^^J
{code}^^J\\end{{lstlisting}}"""

    def block_quote(self, text):
        return text

    def block_html(self, html):
        return html

    def block_error(self, html):
        return html

    def list(self, text, ordered, **attrs):
        if ordered:
            return "\n\\begin{{enumerate}}\n{}\\end{{enumerate}}".format(text)
        else:
            return "\n\\begin{{varwidth}}[t]{{\\linewidth}}\n\\begin{{itemize}}[topsep = 0pt, parsep = 0pt]\n{}\\strut\\end{{itemize}}\end{{varwidth}}\n".format(text)

    def list_item(self, text, level):
        return " \item {}\n".format(text)



from secrets import choice
import string
import logging
import re

log = logging.getLogger(__name__)


def test_color(color):
    """ Return true if color is a hex color value such as a1b2c3"""
    return re.search("[0-9a-fA-F]{6}", color)


class TableStyles:
    def __init__(self):
        self.styles = {}
        self.aliases = {}

    def addstyle(self, name, options):
        """Add a style to the style list.

        Style options can include:
        - alias: 'alias'
        - color: 'color_name'
        - bgcolor: 'color_name'
        - bold: bool
        - italic: bool
        - uppercase: bool
        - alignment: l/left|c/center/centre|r/right
        """

        if 'alias' in options:
            self.aliases[options['alias']] = name

        defaults = {
            'bold': False,
            'italic': False,
            'uppercase': False,
            'alignment': '',
            'color': None,
            'bgcolor': None,
            'size': None
        }

        for default in defaults:
            if default not in options:
                options[default] = defaults[default]

        alignmentaliases = {
            'l': 'l', 'c': 'c', 'r': 'r',
            'left': 'l',
            'center': 'c',
            'centre': 'c',
            'right': 'r'
        }
        if options['alignment'] in alignmentaliases:
            options['alignment'] = alignmentaliases[options['alignment']]
        else:
            options['alignment'] = ''

        options['name'] = name

        self.styles[name] = options

    def addstyles(self, styles):
        for style in styles:
            self.addstyle(style)

    def getstyle(self, name):
        if type(name) is not str:
            return None
        if name in self.styles:
            return self.styles[name]
        if name in self.aliases:
            return self.styles[self.aliases[name]]
        return None

    def getstyles(self):
        return self.styles


class ColorList:
    def __init__(self):
        self.colors = {}

    def addcolor(self, name, color):
        if name != None:
            self.colors[name] = color
            return name
        else:
            # See if it's already there
            for c in self.colors:
                if self.colors[c] == color:
                    return c

            # Give it a random name
            newname = None
            while not newname or newname in self.colors:
                newname = ''.join(
                    [choice(string.ascii_letters) for _ in range(5)])
            self.colors[newname] = color
            return newname

    def addcolors(self, colors):
        for name in colors:
            self.colors[name] = colors[name]

    def getcolor(self, name):
        if name in self.colors:
            return self.colors[name]
        else:
            return None

    def getcolors(self):
        return self.colors



from report_ranger.output_formatter.outputformatter import OutputFormatter
import re
import mistune
from report_ranger.markdown_renderer.csvrenderer import CSVRenderer
from report_ranger.table import Table
import logging
import csv
from tabulate import tabulate

log = logging.getLogger(__name__)

csv_defaults = {
    'columns': {
        'Section': 'section',
        'Name': 'name',
        'Risk': 'risk',
        'Details': 'markdown'
    }
}

header_aliases = {
}


class CSVFormatter(OutputFormatter):
    def __init__(self, templateheaders=dict(), timer=None, watcher=None):
        OutputFormatter.__init__(self, templateheaders, timer, watcher=watcher)
        self.figformat = "png"
        csvoptions = {}
        csvoptions.update(csv_defaults)
        if 'csv' in templateheaders:
            csvoptions.update(templateheaders['csv'])

        self.columns = csvoptions['columns']

    def escape(self, text):
        ''' Escape the given text based on the format we're working with

        :param text: a plain text message
        :return: the message escaped to appear correctly in CSV
        '''
        conv = {
            '\'': r'\'\''
        }
        regex = re.compile('|'.join(re.escape(str(key)) for key in sorted(
            conv.keys(), key=lambda item: - len(item))))
        return regex.sub(lambda match: conv[match.group()], text)

    def table(self, table, options={}, **tableargs):
        ''' This function formats the table in either latex or markdown, depending on the output.'''

        t = Table(table, env=self.env, **tableargs)

        # We're going to do a markdown table with tabulate
        colalign = t.colalign
        # We need to convert the 'l' and 'r' in colalign if it exists
        for col in range(len(colalign)):
            if colalign[col] == 'l':
                colalign[col] = 'left'
            elif colalign[col] == 'c':
                colalign[col] = 'center'
            elif colalign[col] == 'r':
                colalign[col] = 'right'
            else:
                colalign[col] = ''

        if len(table.header) == 0:
            return tabulate(t.table, stralign=colalign, tablefmt="github")
        else:
            return tabulate(t.table[1:], t.table[0], stralign=colalign, tablefmt="github")

    def headers(self):
        return ''

    def output(self, markdown, outputfile=''):
        """ Output vulnerability details to CSV.

        This involves getting the vulnerabilities, splitting the content between vulnerability headings, stripping formatting, and outputting into CSV.
        """

        # Get vulnerability information

        rows = []
        rows.append(list(self.columns.keys()))

        vuln_list = self.env.get('vulnerabilities')
        for section in vuln_list.sections:
            for vuln in section.vulnerabilities:
                lr = CSVRenderer()
                lrm = mistune.create_markdown(renderer=lr)

                potcols = {}
                markdown = vuln.markdown
                mardown_rendered = lrm(markdown)
                potcols['markdown'] = mardown_rendered
                potcols.update(lr.heading_text)
                potcols.update(vuln.headers)
                potcols['section'] = section.name

                row = []
                for col in self.columns:
                    if self.columns[col] in potcols:
                        row.append(potcols[self.columns[col]])
                    else:
                        row.append('')
                rows.append(row)

        log.info("Writing CSV file")
        with open(outputfile, 'w') as fh:
            csvw = csv.writer(fh)
            csvw.writerows(rows)

        return rows



import mistune.renderers.markdown
import re
import logging
from typing import Dict, Any
from mistune.core import BlockState

log = logging.getLogger(__name__)


_strip_end_re = re.compile(r'\n\s+$')

def strip_end(src: str):
    return _strip_end_re.sub('\n', src)

def render_list(renderer, token, state) -> str:
    attrs = token['attrs']
    if attrs['ordered']:
        children = _render_ordered_list(renderer, token, state)
    else:
        children = _render_unordered_list(renderer, token, state)

    text = ''.join(children)
    parent = token.get('parent')
    if parent:
        if parent['tight']:
            return text
        return text + '\n'
    return strip_end(text) + '\n'


def _render_list_item(renderer, parent, item, state):
    leading = parent['leading']
    text = ''
    for tok in item['children']:
        if tok['type'] == 'list':
            tok['parent'] = parent
        elif tok['type'] == 'blank_line':
            continue
        text += renderer.render_token(tok, state)

    lines = text.splitlines()
    text = (lines[0] if lines else '') + '\n'
    prefix = ' ' * len(leading)
    for line in lines[1:]:
        if line:
            text += prefix + line + '\n'
        else:
            text += '\n'
    return leading + text


def _render_ordered_list(renderer, token, state):
    attrs = token['attrs']
    start = attrs.get('start', 1)
    for item in token['children']:
        leading = str(start) + token['bullet'] + ' '
        parent = {
            'leading': leading,
            'tight': token['tight'],
        }
        yield _render_list_item(renderer, parent, item, state)
        start += 1


def _render_unordered_list(renderer, token, state):
    parent = {
        'leading': '-' + ' ',
        'tight': token['tight'],
    }
    for item in token['children']:
        yield _render_list_item(renderer, parent, item, state)


class TypstRenderer(mistune.renderers.markdown.MarkdownRenderer):
    NAME = 'typst'
    IS_TREE = False
    ordered = False
    escape_text = False

    list_depth=1

    def escape(self, text):
        ''' Escape the given text for display in Typst output

        :param text: a plain text message
        :return: the message escaped to appear correctly in Typst
        '''
        conv = {
            '&': r'\&',
            '%': r'\%',
            '$': r'\$',
            '#': r'\#',
            '_': r'\_',
            '{': r'\{',
            '}': r'\}',
            '[': r'\[',
            ']': r'\]',
            '@': r'\@',
            '*': r'\*',
            '\\': r'\\',
            '<':'\<',
            '>':'\>'
        }
        regex = re.compile('|'.join(re.escape(str(key)) for key in sorted(
            conv.keys(), key=lambda item: - len(item))))
        return regex.sub(lambda match: conv[match.group()], text)



    def __init__(self, escape=True):
        super(TypstRenderer, self).__init__()
        self.escape_text = escape

    def text(self, token: Dict[str, Any], state: BlockState) -> str:
        return self.escape(token['raw'])

    def link(self, token: Dict[str, Any], state: BlockState) -> str:
        label = token.get('label')
        text = self.render_children(token, state)

        attrs = token['attrs']
        url = attrs['url']
        title = attrs.get('title')

        if title:
            return f'#link("{url}")'
        else:
            return f'#link("{url}")[{text}]'

    def image(self, token: Dict[str, Any], state: BlockState) -> str:
        label = token.get('label')
        text = self.render_children(token, state)

        attrs = token['attrs']
        url = attrs['url']
        title = attrs.get('title')

        if title:
            return f'#figure(image("{url}"), caption: [{text}])'
        else:
            return f'#image("{url}")'
        

    def emphasis(self, token: Dict[str, Any], state: BlockState) -> str:
        return '_' + self.render_children(token, state) + '_'

    def strong(self, token: Dict[str, Any], state: BlockState) -> str:
        return '*' + self.render_children(token, state) + '*'
    
    def thematic_break(self, token: Dict[str, Any], state: BlockState) -> str:
        return '\n\n'

    def heading(self, token: Dict[str, Any], state: BlockState) -> str:
        level = token['attrs']['level']
        marker = '=' * level
        text = self.render_children(token, state)
        if text[-6:] == ' \{-\}':
            text = text[:-6]
        return marker + ' ' + text + '\n\n'
    
    def list(self, token: Dict[str, Any], state: BlockState) -> str:
        return render_list(self, token, state)




import logging

log = logging.getLogger(__name__)

def log_jinja2_error(markdown, error):
    log.error("Affected lines:")
    mdlines = markdown.splitlines()
    for i in range(5):
        el = error.lineno - 5 + i
        if el < 0:
            continue
        log.error("{}: {}".format(el, mdlines[el]))
    log.error("+ {}: {}".format(error.lineno, mdlines[error.lineno]))
    for i in range(5):
        el = error.lineno + 1 + i
        if el >= len(mdlines):
            continue
        log.error("{}: {}".format(el, mdlines[el]))


from uuid import uuid4
import re
import logging

log = logging.getLogger(__name__)


class ContentAssistant:
    def __init__(self):
        self.defaulticounter = None
        self.icounters = {}
        self.defaultcounter = 0
        self.counters = {}
        self.defaulttable = []
        self.tables = {}
        self._register = {}
        self._string_register = {}
        self._table_register = {}  # For holding table formatting
        self._function_register = {} # For holding simple function calls

    def _display_table(self, uuid, markdown, of, tablename="", *args, **kwargs):
        """ Display a table from
        """
        if tablename == "":
            table = self.defaulttable
        elif not isinstance(tablename, str):
            log.warn(
                "CA table name you're trying to display to is not a string, skipping: {}".format(tablename))
            return ""
        elif tablename not in self.tables:
            log.warning(
                "Table {} has not had any rows added to it. Skipping the display of this table.".format(tablename))
            return ""
        else:
            table = self.tables[tablename]

        return markdown.replace(uuid, of.table(table, *args, **kwargs))

    def _display_icounter(self, uuid, markdown, of):
        """ Convert the counter UUIDs in the document to numbers.
        """
        searchre = re.compile("(x|y|z){}".format(uuid))
        counter = 0
        max = 10000
        while max > 0:
            max -= 1
            match = searchre.search(markdown)
            if match == None:
                return markdown
            if(match.group()[0]) == 'x':
                counter += 1
            if(match.group()[0]) == 'z':
                counter = 0
                markdown = searchre.sub("", markdown, 1)
            else:
                markdown = searchre.sub(str(counter), markdown, 1)
        log.warn("Max counter size of 10000 reached.")
        return markdown
    
    def _register_string(self, string):
        newuuid = str(uuid4())
        self._string_register[newuuid] = string
        return newuuid

    def _register_ca(self, function, *args, **kwargs):
        """ Add a function to the register. This can be called later so that we can display content on the second pass
        """
        newuuid = str(uuid4())
        self._register[newuuid] = {
            'function': function, 'args': args, 'kwargs': kwargs}
        return newuuid

    def _register_formattable(self, *args, **kwargs):
        """ Add a function to the register. This can be called later so that we can display content on the second pass
        """
        newuuid = str(uuid4())
        self._table_register[newuuid] = {'args': args, 'kwargs': kwargs}
        return newuuid

    def parse_register(self, of, markdown):
        """ Go through the register and print out the results of the functions
        """
        log.info("Adding Content Assistant content")

        # CA register
        for uuid, f in self._register.items():
            markdown = f['function'](
                uuid, markdown, of, *f['args'], **f['kwargs'])

        # Now to go through calls to ft(). The below code will replace the markdown table with a table() call.
        t_r_row = r'\|?([^\|\r\n]+\|)+[^\|\r\n]*'
        t_r_header_sep = r'\|?(\:?\s+-+\s*\:?\s*\|)*'
        table_regex = r'[\r\s\n]*((' + t_r_row + \
            r')?[\r\s\n]*(' + t_r_header_sep + \
            r')?([\r\s\n]*' + t_r_row + ')+)'

        # Table register
        for uuid, f in self._table_register.items():
            m = re.search(uuid + table_regex, markdown)
            if not m:
                log.warn(
                    "Table formatting string does not seem to be attached to a markdown table")
                markdown = markdown.replace(uuid, '')
                continue
            table_markdown = m.group(1)
            table_output = of.table(table_markdown, *f['args'], **f['kwargs'])
            start, end = m.span()
            markdown = markdown[:start] + \
                table_output + markdown[end:]

        # String register
        if len(self._string_register) > 0:
            regex = re.compile('|'.join(re.escape(str(key)) for key in sorted(
                self._string_register.keys(), key=lambda item: - len(item))))
            markdown = regex.sub(lambda match: self._string_register[match.group()], markdown)

        return markdown

    def icounter(self, countername="", mode="x"):
        """ This function allows counters throughout the document. Multiple counters can be used by specifying counter names.
        """
        if mode not in ['x', 'y', 'z']:
            log.warn("Unknown mode for icounter {}".format(countername))
            return ""

        if countername == "":
            if self.defaulticounter == None:
                self.defaulticounter = self._register_ca(
                    self._display_icounter)
            return "{}{}".format(mode, self.defaulticounter)

        elif not isinstance(countername, str):
            log.warn("Counter name not a string: {}".format(countername))
            return 0
        else:
            if countername not in self.icounters:
                self.icounters[countername] = self._register_ca(
                    self._display_icounter)
            return "{}{}".format(mode, self.icounters[countername])

    def get_icounter(self, countername=""):
        """ This function retrieves the value of the counter WITHOUT iterating it.
        """
        return self.icounter(countername, "y")

    def reset_icounter(self, countername=""):
        return self.icounter(countername, "z")

    def counter(self, countername="", iterate=True):
        """ This function allows counters throughout the document. Multiple counters can be used by specifying counter names.
        """

        if countername == "":
            if iterate:
                self.defaultcounter += 1
            return self.defaultcounter
        elif not isinstance(countername, str):
            log.warn("Counter name not a string: {}".format(countername))
            return 0
        else:
            if countername not in self.counters:
                self.counters[countername] = 0
            if iterate:
                self.counters[countername] += 1
            return self.counters[countername]

    def get_counter(self, countername=""):
        """ This function retrieves the value of the counter WITHOUT iterating it.
        """
        return self.counter(countername, False)

    def reset_counter(self, countername=""):
        """ Reset a counter to 0
        """
        if countername == "":
            self.defaultcounter = 0
        elif not isinstance(countername, str):
            log.warn(
                "Counter name you're trying to reset is not a string, skipping: {}".format(countername))
        else:
            self.counters[countername] = 0
        return ""

    def table_row(self, tablename="", row=[]):
        """ Add a row to a content assistant table, which can then be displayed with display_table
        """
        if tablename == "":
            self.defaulttable.append(row)
        elif not isinstance(tablename, str):
            log.warn(
                "CA table name you're trying to add a row to is not a string, skipping: {}".format(tablename))
        else:
            if tablename not in self.tables:
                self.tables[tablename] = []
            self.tables[tablename].append(row)
        return ""

    def table_rows(self, tablename="", rows=[]):
        """ Add multiple rows to a content assistant table
        """
        if tablename == "":
            self.defaulttable += rows
        elif not isinstance(tablename, str):
            log.warn(
                "CA table name you're trying to add rows to is not a string, skipping: {}".format(tablename))
        else:
            if tablename not in self.tables:
                self.tables[tablename] = []
            self.tables[tablename] += rows
        return ""

    def display_table(self, *args, **kwargs):
        return self._register_ca(self._display_table, *args, **kwargs)

    def format_table(self, *args, **kwargs):
        """ Format the markdown table that's below. This allows simpler syntax for fancy formatting of markdown tables, all in jinja2 templating.
        It requires two passes, with this function returning a UUID and the second pass grepping the UUID and markdown to call of.table.
        """
        return self._register_formattable(*args, **kwargs)



class InputError(Exception):
    """Exception raised for errors in the input.

    Attributes:
        error -- explanation of the error
    """

    def __init__(self, message):
        self.message = message


import os
from platformdirs import user_config_dir
import yaml
import logging
import copy
import report_ranger

log = logging.getLogger(__name__)


# This is to fill in the defaults so there's not too many variables
config = {
    "default_template": '',
    "default_input_file": 'reportbody.md',
    "default_output_file": 'report-preview.md',
    "format": '',
    "verbose": False,
    # The template mapper. This gives the locations of template files for each template.
    "templatemapper": {
    },
    # Additional template mapper files to link.
    "templatemappers": [
    ],
    # Files with additional
    "includes": {
    }
}

envvars = {
    'RR_TEMPLATE': "default_template",
    'RR_INPUT_FILE': 'default_input_file',
    'RR_OUTPUT_FILE': 'default_input_file',
    'RR_VERBOSE': 'verbose',
    'RR_TEMPLATEMAPPERS': 'templatemappers'
}

def get_config_from_file(file):
    try:
        with open(file, 'r') as vf:
            config = yaml.safe_load(vf)
            log.debug(f"Processing config file {file}")
            return config
    except:
        log.debug(f"Could not open config file {file}")
        return {}

def get_config(arg_file = None):
    final_config = copy.copy(config)
    
    for config_location in [user_config_dir('reportranger', 'Volkis') + '/config.yml', 'config.yml']:
        if config_location:
            final_config.update(get_config_from_file(config_location))

    final_config['templatemapper']["sample"] = os.path.join(os.path.dirname(os.path.realpath(report_ranger.__file__)),"default-templates/sample-template/rr-sample-template.md") # Add in the sample template
    final_config['templatemapper']["plain"] = os.path.join(os.path.dirname(os.path.realpath(report_ranger.__file__)),"default-templates/plain-template/rr-plain-template.md") # Add in the sample template
    
    for config_location in [user_config_dir('reportranger', 'Volkis') + '/config.yml', 'config.yml']:
        if config_location:
            final_config.update(get_config_from_file(config_location))

    # Sort out environment variables
    for envvar in envvars.keys():
        log.debug(f"Getting envvar {envvar}")
        var = os.getenv(envvar)
        log.debug(f"Got {var}")
        if var != None:
            log.debug(f"Setting config variable {envvars[envvar]} to {var} from environment variable {envvar}")
            if isinstance(final_config[envvars[envvar]], dict):
                final_config[envvars[envvar]] += var.split(':')
            final_config[envvars[envvar]] = var
    
    for config_location in [arg_file]:
        if config_location:
            final_config.update(get_config_from_file(config_location))
    
    return final_config




from report_ranger.output_formatter.outputformatter import OutputFormatter, headeralias
import re
import os
import yaml
import jinja2
import mistune
from report_ranger.markdown_renderer.latexrenderer import LatexTableRenderer
from report_ranger.table import Table
import subprocess
import logging
from report_ranger.utils.mdread import process_template


log = logging.getLogger(__name__)

default_latex = {'title': '{{title}}',
                 'author': '[Author Name]',
                 'date': '{{date.strftime("%-d %B %Y")}}',
                 'subtitle': 'Prepared for {% if longclient %}{{longclient}}{% else %}{{client}}{% endif %}',
                 'template': '',
                 'lang': 'en',
                 'twoside': False,
                 'numbersections': False,
                 'titlepage': True,
                 'titlepage-text-color': '000000',
                 'titlepage-rule-height': 0,
                 'titlepage-background': 'titlepage.pdf',
                 'page-background': '',
                 'page-background-opacity': 1.0,
                 'section-background': '',
                 'section-skip': '0pt',
                 'toc': True,
                 'toc-title': 'Table of Contents',
                 'toc-own-page': True,
                 'toc-depth': 2,
                 'color-table-header': 'C72727',
                 'color-table-1': 'FFFFFF',
                 'color-table-2': 'FFFFFF',
                 'color-informational': '3366FF',
                 'color-low': 'D3EAF9',
                 'color-medium': 'FFFF00',
                 'color-high': 'FF0000',
                 'color-high-font': '000000',
                 'color-critical': '000000',
                 'color-open': 'E47676',
                 'color-open-font': '000000',
                 'color-closed': 'A9DA74',
                 'color-closed-font': '000000',
                 'color-theme': 'C72727',
                 'toc-section': True,
                 'table-styles': ''}

header_aliases = {
    'titlepage-text-colour': 'titlepage-text-color',
    'titlepage-rule-colour': 'titlepage-rule-color',
    'colour-table-header': 'color-table-header',
    'colour-table-1': 'color-table-1',
    'colour-table-2': 'color-table-2',
    'colour-informational': 'color-informational',
    'colour-informational-font': 'color-informational-font',
    'colour-low': 'color-low',
    'colour-low-font': 'color-low-font',
    'colour-medium': 'color-medium',
    'colour-medium-font': 'color-medium-font',
    'colour-high': 'color-high',
    'colour-high-font': 'color-high-font',
    'colour-critical': 'color-critical',
    'colour-critical-font': 'color-critical-font',
    'colour-theme': 'color-theme'
}

latex_default_table_options = {
    'fullwidth': True,
    'vlines': True,
    'hlines': True,
    'color-1': 'color-table-1',
    'color-2': 'color-table-2',
    'banding': True,
    'longtable': True,
    'rowhead': 1
}


def stripmdpara(text):
    if text[-4:] == ' \\\\ ':
        return text[:-4]
    else:
        return text


class LatexFormatter(OutputFormatter):
    def __init__(self, templateheaders=dict(), timer=None, watcher=None):
        OutputFormatter.__init__(self, templateheaders, timer, watcher=watcher)
        self.templatefile = self.templateheaders['latex_template']
        self.figformat = "pdf"

        # Set up latex file headers
        latexheaders = dict()

        # Put in the defaults for the latex template
        latexheaders.update(default_latex)

        # Update latex headers with what's in the template
        if "latex" in templateheaders:
            for header in header_aliases.keys():
                if header in templateheaders['latex']:
                    templateheaders['latex'][header_aliases[header]
                                           ] = templateheaders['latex'][header]
                    del templateheaders['latex'][header]

            latexheaders.update(templateheaders['latex'])

        if ('header-left' in latexheaders and latexheaders['header-left'] != '') or ('header-center' in latexheaders and latexheaders['header-center'] != '') or ('header-right' in latexheaders and latexheaders['header-right'] != ''):
            latexheaders['has-header'] = 'true'

        # Add the toc-section if there's a section
        if latexheaders['toc-section']:
            latexheaders['toc-section'] = self.newsection()
        else:
            latexheaders['toc-section'] = False

        # Add over the headers that refer to local files
        for header in ['titlepage-background', 'page-background']:
            if header in latexheaders:
                latexheaders[header] = os.path.join(
                    self.templateheaders['templatedir'], latexheaders[header])

        # Set the default table options
        self.default_table_options = latex_default_table_options

        colors = ''
        colorlist = self.colors.getcolors()
        for name, color in colorlist.items():
            colors += '\\definecolor{{rrcolor{}}}{{HTML}}{{ {} }}\n'.format(
                name, color)
        if colors != '':
            latexheaders['colors'] = f'{colors}\n'

        # Address if there are custom table styles
        if not latexheaders['table-styles']:
            tablestyles = ''
            ts = self.tablestyles.getstyles()
            for name, style in ts.items():
                log.debug("Adding style {}".format(name))

                textcolor = ''
                if style['color']:
                    textcolor = "\\textcolor{{rrcolor{}}}".format(
                        style['color'])

                cellcolor = ''
                if style['bgcolor']:
                    cellcolor = '\\cellcolor{{rrcolor{}}} '.format(
                        style['bgcolor'])
                bold = ''
                if style['bold'] == True:
                    bold = '\\bfseries '

                if style['size']:
                    s = style['size']
                    sizes = {
                        'tiny': '\\tiny',
                        'small': '\\small',
                        'normal': '',
                        'large': '\\large',
                        'Large': '\\Large',
                        'LARGE': '\\LARGE',
                        'huge': '\\huge',
                        'Huge': '\\Huge'
                    }
                    if s in sizes:
                        size = sizes[s]
                else:
                    size = ''

                if style['uppercase']:
                    uppercase = '\\MakeUppercase{{{} #1 }}'.format(size)
                else:
                    uppercase = '{} #1'.format(size)

                tablestyles += "\\newcommand*{{\\rrts{}}}[1]{{ {{ {} {} {{ {} {} }} }} }}\n".format(
                    name,
                    cellcolor,
                    textcolor,
                    bold,
                    uppercase
                )
            latexheaders['table-styles'] = tablestyles
        else:
            log.debug("NO TABLE STYLES: {}".format(
                latexheaders['table-styles']))
            log.debug("NO TABLE STYLES: {}".format(
                latexheaders))

        if 'default-table-options' in latexheaders:
            if type(latexheaders['default-table-options']) is dict:
                # Overlay the template defaults
                self.default_table_options.update(
                    latexheaders['default-table-options'])
            else:
                log.warn(
                    "default-table-options variable in the latex headers is not a dict, ignoring")

        self.latexheaders = latexheaders

    def escape(self, text):
        ''' Escape the given text based on the format we're working with

        :param text: a plain text message
        :return: the message escaped to appear correctly in LaTeX
        '''
        if type(text) is not str:
            log.warning(
                "escape function given {} which is not a string. Returning ''".format(text))
            return ""

        conv = {
            '&': r'\&',
            '%': r'\%',
            '$': r'\$',
            '#': r'\#',
            '_': r'\_',
            '{': r'\{',
            '}': r'\}',
            '~': r'\textasciitilde{}',
            '^': r'\^{}',
            '\\': r'\textbackslash{}',
            '<': r'\textless{}',
            '>': r'\textgreater{}',
        }
        regex = re.compile('|'.join(re.escape(str(key)) for key in sorted(
            conv.keys(), key=lambda item: - len(item))))
        return regex.sub(lambda match: conv[match.group()], text)

    def table(self, table, options={}, **tableargs):
        ''' This function formats the table in either latex or markdown, depending on the output. '''
        markdown = ""

        opts = self.default_table_options.copy()
        opts.update(options)

        t = Table(table, env=self.env, **tableargs)

        toptions = []
        if opts['vlines']:
            if 'vlines-options' in opts:
                toptions.append(f"vlines = {opts['vlines-options']}")
            else:
                toptions.append("vlines")
                
        if opts['hlines']:
            if 'hlines-options' in opts:
                toptions.append(f"hlines = {opts['hlines-options']}")
            else:
                toptions.append("hlines")
        if opts['rowhead'] > 0 and len(t.table) > 1:
            toptions.append("rowhead = {}".format(opts['rowhead']))

        # If there's no actual table, warn and exit
        if t.width == 0:
            log.warn("Table is now empty. Printing an empty string.")
            return ""

        lr = mistune.create_markdown(renderer=LatexTableRenderer())

        if opts['banding']:
            toptions.append("row{{odd}} = {{bg={}}}".format(opts['color-1']))
            toptions.append("row{{even}} = {{bg={}}}".format(opts['color-2']))

        colspec = []
        for c in range(len(t.table[0])):
            col = ''
            if opts['fullwidth']:

                copts = []

                if tableargs.get('colwidths'):
                    if t.colwidths[int(c)] == 0:
                        col += 'Q'
                    else:
                        col += 'X'
                        copts.append(str(t.colwidths[int(c)]))
                else:
                    col += 'X'
                if t.colalign[c] != "":
                    copts.append(t.colalign[c])

                if len(copts) > 0:
                    col += "[{}]".format(",".join(copts))
            else:
                if t.colalign[c] != "":
                    col = t.colalign[c]
                else:
                    col = 'l'

            colspec.append(col)

        toptions.append("colspec={{{}}}".format("".join(colspec)))
        if opts['longtable']:
            # Hack to remove header caption
            if 'headerhack' in opts and opts['headerhack']:
                markdown += "\\vspace{{-1.5\\baselineskip}}"
            markdown += "\\SetTblrTemplate{{head}}{{empty}}".format()
            markdown += "\\begin{{longtblr}}{{ {} }} \n".format(
                ",".join(toptions))
        else:
            markdown += "\\begin{{tblr}}{{ {} }} \n".format(",".join(toptions))

        for i in range(len(t.table)):
            for j in range(len(t.table[i])):
                style = self.tablestyles.getstyle(t.cellstyles[i][j])
                if j > 0:
                    markdown += ' & '

                if t.colspan[i][j] < 1:  # This column is being overwritten in some way
                    continue
                if t.rowspan[i][j] < 1:  # This column is being overwritten in some way
                    continue

                mr = []
                # Multi-column
                if t.colspan[i][j] > 1:
                    mr.append("c={}".format(t.colspan[i][j]))

                # Multi-row
                if t.rowspan[i][j] > 1:
                    mr.append("r={}".format(t.rowspan[i][j]))

                mrspec = "" if len(mr) == 0 else "[{}]".format(",".join(mr))

                cellopts = []

                # cell alignment, first try cellalign, then style
                if t.cellalign[i][j] != '':
                    cellopts.append("halign={}".format(t.cellalign[i][j]))
                elif style and style['alignment'] != '':
                    cellopts.append("halign={}".format(style['alignment']))

                cellinit = ''
                if style:
                    if style['color']:
                        cellopts.append("fg=rrcolor{}".format(style['color']))
                    if style['bgcolor']:
                        cellopts.append(
                            "bg=rrcolor{}".format(style['bgcolor']))
                    if style['bold']:
                        cellinit += "\\bfseries "
                    if style['italic']:
                        cellinit += "\\itshape "

                    if style['size']:
                        s = style['size']
                        sizes = {
                            'tiny': '\\tiny',
                            'small': '\\small',
                            'normal': '',
                            'large': '\\large',
                            'Large': '\\Large',
                            'LARGE': '\\LARGE',
                            'huge': '\\huge',
                            'Huge': '\\Huge'
                        }
                        if s in sizes:
                            cellinit += sizes[s]
                elif t.cellstyles[i][j] != '':
                    if t.cellstyles[i][j] in headeralias:
                        t.cellstyles[i][j] = headeralias[t.cellstyles[i][j]]

                if len(mr) > 0 or len(cellopts) > 0:
                    markdown += "\\SetCell{}{{{}}}".format(
                        mrspec, ",".join(cellopts))

                cell = t.table[i][j]

                # If we've got a split box
                if(isinstance(cell, list)) and not isinstance(cell, str):
                    if len(cell) > 1:
                        markdown += "\\diagbox{{{} {} }}{{{} {} }}".format(cellinit, stripmdpara(lr(
                            str(cell[0]))), cellinit, stripmdpara(lr(str(cell[1]))))
                    else:
                        markdown += "{{{} {} }}".format(cellinit,
                                                        stripmdpara(lr(str(cell[0]))))
                else:
                    celltext = stripmdpara(lr(str(cell)))
                    markdown += "{{{} {} }}".format(cellinit, celltext)

            markdown += " \\\\ \n"

        if opts['longtable']:
            markdown += "\\end{longtblr}\n"
        else:
            markdown += "\\end{tblr}\n"
        return markdown

    def newsection(self):
        if "newsection-content" in self.templateheaders['latex']:
            output = process_template(
                [], self.templateheaders['latex']['newsection-content'], env=self.env, name="section", filename="")
            return output
        markdown = '\\newpage\n'
        if "section-background" in self.templateheaders['latex']:
            markdown += '\\begin{tikzpicture}[remember picture,overlay]\n'
            markdown += '\\node[inner sep=0] at (current page.center)\n'
            markdown += '{\\includegraphics[width=\\paperwidth,height=\\paperheight]{' + os.path.join(
                self.templateheaders['templatedir'], self.templateheaders['latex']["section-background"]) + '}};\n'
            markdown += '\\end{tikzpicture}\n'
            if "section-skip" in self.templateheaders['latex']:
                markdown += '\\vskip {}\n'.format(
                    self.templateheaders['latex']['section-skip'])
            else:
                markdown += '\\vskip 14em\n'
        return markdown

    def newpage(self):
        return "\\pagebreak"
    
    def _process_template_headers(self, header, headername, env):
        if type(header) == str:
            # Run a jinja template if it's a string
            try:
                j2template = jinja2.Template(header)
                return j2template.render(env)
            except jinja2.exceptions.TemplateSyntaxError as error:
                log.error(f"Error in processing template header {headername}: {error.message}")
                log.error(f"Header contents: {header}")
                log.error(f"Returning unprocessed header")
                return header
            except jinja2.exceptions.TemplateError as error:
                log.error(f"Error in processing template header {headername}: {error.message}")
                log.error(f"Returning unprocessed header")
                return header
        if type(header) == list:
            # If it's a list, rerun on each element of the list
            return [self._process_template_headers(headervalue, headername, env) for headername, headervalue in enumerate(header)]
        if type(header) == dict:
            # If it's a dict, rerun on each element of the dict
            return {headername: self._process_template_headers(headervalue, headername, env) for headername, headervalue in header.items()}
        # If it's anything else (number, date, etc), just return it blank.
        return header
                



    def headers(self):
        
        env = self.env.get_env()

        markdown = yaml.dump(self._process_template_headers(self.latexheaders, "headers", env))
        return markdown

    def end(self):
        if "endpage-content" in self.templateheaders['latex']:
            output = process_template(
                [], self.templateheaders['latex']['endpage-content'], env=self.env, name="endpage", filename="")
            return output
        else:
            markdown = '\\newpage\n'
            """ To be called at the end of the output - puts in the ending page/graphics """
            markdown = ''
            markdown += '\\pagebreak\n'
            markdown += '\\thispagestyle{empty}\n'
            markdown += '\\NoBgThispage\n'
            markdown += '\\begin{tikzpicture}[remember picture,overlay]\n'
            markdown += '\\node[inner sep=0] at (current page.center)\n'
            markdown += '{\\includegraphics[width=\\paperwidth,height=\\paperheight]{' + os.path.join(
                self.templateheaders['templatedir'], self.templateheaders['latex']['titlepage-background']) + '}};\n'
            markdown += '\\end{tikzpicture}\n'
            markdown += '\\vskip 14em\n'
            return markdown

    def output(self, markdown, outputfile=''):
        output = self._build_markdown(markdown)

        if self.docformat and self.docformat != 'pdf':
            log.debug("Writing Markdown")
            with open(outputfile, 'w') as fh:
                fh.write(output)
            log.info("Finished writing Markdown")

            return output

        log.debug("Writing PDF")
        log.debug(f"Latex template file: {self.templatefile}")
        log.debug(f"Output file: {outputfile}")
        # Use Pandoc to print to PDF
        pandoc_arguments = ['pandoc', '--from', 'markdown', '--to', 'latex',
                            '--template', self.templatefile, '--listings', '-o', outputfile]
        if self.options and 'latex_numbered-headers' in self.options and self.options['latex_numbered-headers']:
            pandoc_arguments.append('--number-sections')
        process = subprocess.run(pandoc_arguments,
                                 input=output,
                                 universal_newlines=True)
        
        log.info("Finished writing PDF file")

        return output



import os
from pathlib import Path
import shutil
import subprocess
from report_ranger.output_formatter.outputformatter import OutputFormatter, headeralias
import re
import jinja2
import mistune
from report_ranger.markdown_renderer.typstrenderer import TypstRenderer
from report_ranger.table import Table
import logging
from report_ranger.utils.jinja_helpers import log_jinja2_error
from report_ranger.utils.mdread import process_template


log = logging.getLogger(__name__)

class TypstFormatter(OutputFormatter):
    def __init__(self, templateheaders=dict(), timer=None, watcher=None):
        OutputFormatter.__init__(self, templateheaders, timer, watcher=watcher)
        self.figformat = "svg"
        self.env.set_static('table', self._register_table)
        log.debug(f"Typst init watcher: {self.watcher}")

    def escape(self, text):
        ''' Escape the given text based on the format we're working with

        :param text: a plain text message
        :return: the message escaped to appear correctly in Typst
        '''
        if type(text) is not str:
            log.warning(
                "escape function given {} which is not a string. Returning ''".format(text))
            return ""

        conv = {
            '&': r'\&',
            '"': r'\\"',
            '\'': r'\\\'',
            '%': r'\%',
            '$': r'\$',
            '#': r'\#',
            '_': r'\_',
            '{': r'\{',
            '}': r'\}',
            '[': r'\[',
            ']': r'\]',
            '@': r'\@',
            '\\': r'\\'
        }
        regex = re.compile('|'.join(re.escape(str(key)) for key in sorted(
            conv.keys(), key=lambda item: - len(item))))
        return regex.sub(lambda match: conv[match.group()], text)
    
    def _register_table(self, *args, **kwargs):
        ca = self.env.get('ca')
        return ca._register_string(self.table(*args, **kwargs))

    def _translate_alignment(self, alignment):
        # Translates from 'l','c','r','j' to typst
        if alignment == 'l':
            return 'left'
        if alignment == 'r':
            return 'right'
        if alignment == 'c':
            return 'center'
        return 'left'

    def table(self, table, options={}, **tableargs):
        ''' This function formats the table in either latex or markdown, depending on the output. '''
        typst = ""

        t = Table(table, env=self.env, **tableargs)

        # If there's no actual table, warn and exit
        if t.width == 0:
            log.warn("Table is now empty. Printing an empty string.")
            return ""

        #TODO: Add banding

        columndef = ",".join(["auto" if cw == 0 else f"{cw}fr" for cw in t.colwidths])

        cells = []
        colors = self.colors.getcolors()

        for i in range(len(t.table)):
            for j in range(len(t.table[i])):
                parameters = []
                textparameters = []

                lr = mistune.create_markdown(renderer=TypstRenderer())

                # Process the typst text
                cell = lr(str(t.table[i][j]))

                style = self.tablestyles.getstyle(t.cellstyles[i][j])

                if t.colspan[i][j] < 1:  # This column is being overwritten in some way
                    continue
                elif t.colspan[i][j] > 1:
                    parameters.append(f"colspan: {t.colspan[i][j]}")
                if t.rowspan[i][j] < 1:  # This column is being overwritten in some way
                    continue
                elif t.rowspan[i][j] > 1:
                    parameters.append(f"rowspan: {t.rowspan[i][j]}")
                
                # cell alignment, first try cellalign, then style
                if t.cellalign[i][j] != '':
                    parameters.append(f"align: {self._translate_alignment(t.cellalign[i][j])}")
                elif style and style['alignment'] != '':
                    parameters.append(f"align: {self._translate_alignment(style['alignment'])}")
                elif len(t.colalign) > j and t.colalign[j] != '':
                    parameters.append(f"align: {self._translate_alignment(t.colalign[j])}")
                
                if style:
                    if style['color']:
                        color = style['color']
                        if color in colors:
                            color = colors[color]
                        textparameters.append(f'fill: rgb("{color}")')
                    if style['bgcolor']:
                        color = style['bgcolor']
                        if color in colors:
                            color = colors[color]
                        parameters.append(
                            f'fill: rgb("{color}")')
                    if style['bold']:
                        textparameters.append('weight: "bold"')
                    if style['italic']:
                        textparameters.append('style: "italic"')
                    if style['size']:
                        s = style['size']
                        sizes = {
                            'tiny': '0.5em',
                            'small': '0.8em',
                            'normal': '',
                            'large': '1.3em',
                            'Large': '1.7em',
                            'LARGE': '2em',
                            'huge': '3em',
                            'Huge': '4em'
                        }
                        if s in sizes:
                            textparameters.append(f'size: {sizes[s]}')

                if len(textparameters) > 0:
                    cell = f"#text({','.join(textparameters)})[{cell}]"
                if len(parameters) > 0:
                    cell = f"table.cell({','.join(parameters)})[{cell}]"
                else:
                    cell = f"[{cell}]"
                
                cells.append(cell)

        tablespan = ",\n".join(cells)
        
        typst = f'#table(columns: ({columndef}), {tablespan})'

        return typst

    def newsection(self):
        ca = self.env.get('ca')
        return ca._register_string("#new_section")

    def newpage(self):
        ca = self.env.get('ca')
        return ca._register_string("#pagebreak(weak: true)")
    
    def _build_markdown_headers(self, headers):
        return ''

    def headers(self):
        return ''

    def end(self):
        ca = self.env.get('ca')
        return ca._register_string("#last_page")

    def _build_markdown(self, templatemarkdown):
        # A giant string to put all the output markdown into
        markdown = ""

        log.debug("Outputting markdown")
        env = self.env.get_env()

        try:
            j2template = jinja2.Template(templatemarkdown)
            processedtemplate = j2template.render(env)
            markdown += processedtemplate

            self._warn_in_text(self.templateheaders, env, markdown)


        except jinja2.exceptions.TemplateSyntaxError as error:
            log.error(
                "Error in processing the final template: {}".format(error.message))
            log_jinja2_error(markdown, error)
            raise Exception("Error reading the final template: {} at lineno {} for file {}".format(
                error.message, error.lineno, error.filename))
        except jinja2.exceptions.TemplateError as error:
            log.error(
                "Error in processing the final template: {}".format(error.message))
            raise Exception("Error reading the final template: {}".format(
                error.message))

        return markdown
    
    def _join_path(self, root_dir, dir_name) -> Path:
        root_dir = Path(root_dir)
        joined_dir = root_dir / dir_name
        if root_dir in joined_dir.parents:
            return joined_dir
        raise Exception(f"{joined_dir} not a subdirectory of {root_dir}")
    
    def _copy_datadir(self, dir_name, dest='typst_template'):
        try:
            data_dir:Path = self._join_path(self.templateheaders['templatedir'], dir_name)
            dest_dir:Path = self._join_path(os.path.curdir, dest)
            
            if not dest_dir.exists():
                log.info(f"Copying Typst data directory '{data_dir}' to '{dest_dir}'")
                shutil.copytree(data_dir, dest_dir)
        except Exception as e:
            log.warn(f"Error copying Typst data directory: {e.args}")

    def _output_typst_var(self, variable):
        if type(variable) == str:
            return f'"{self.escape(variable)}"'
        if type(variable) == dict:
            vars = []
            for key, var in dict.items():
                new_key = re.sub(r'[^a-zA-Z0-9]', '', key)
                vars.append(f'{new_key}:{self._output_typst_var(var)}')
            return f"({','.join(vars)})"
        if type(variable) == list:
            vars = []
            for var in variable:
                vars.append(self._output_typst_var(var))
            return f"({','.join(vars)})"
        else:
            return f'"{str(variable)}"'
                            
    def output(self, markdown, outputfile=''):
        output = self._build_markdown(markdown)

        lr = mistune.create_markdown(renderer=TypstRenderer())

        # Process the typst text
        output = lr(output)

        # Get the template and data directory
        if 'typst' in self.templateheaders:
            if 'template' in self.templateheaders['typst']:
                templatefile = os.path.join(
                    self.templateheaders['templatedir'], self.templateheaders['typst']['template'])
                with open(templatefile) as tf:
                    output = tf.read() + output
            if 'data_dir' in self.templateheaders['typst']:
                data_dir = self.templateheaders['typst']['data_dir']
                log.info(f"Copying data dir {data_dir}")
                self._copy_datadir(data_dir, 'typst_template')
            if 'libs_dir' in self.templateheaders['typst']:
                self._copy_datadir(self.templateheaders['typst']['libs_dir'], 'libs')
            if 'template_variables' in self.templateheaders['typst']:
                # Add in env
                docenv = self.env.get_env()
                for var in self.templateheaders['typst']['template_variables']:
                    if var in docenv.keys():
                        output = f'#let {var} = {self._output_typst_var(docenv[var])}\n' + output
                    else:
                        output = f'#let {var} = none\n' + output


        # Content assistant parsing
        output = self.env.get('ca').parse_register(self, output)

        if self.docformat and self.docformat == 'pdf' and outputfile[-4:].lower() == '.pdf':
            outputfilebase = outputfile[:-4]
        elif outputfile[-4:].lower() == '.typ':
            outputfilebase = outputfile[:-4]
        else:
            outputfilebase = outputfile

        log.debug("Writing Typst")
        with open(f"{outputfilebase}.typ", 'w') as fh:
            fh.write(output)
            log.info("Finished writing Typst file")

        if self.docformat and self.docformat == 'pdf':
            log.debug("Writing PDF")
            log.debug(f"Typst file: {outputfilebase}.typ")
            # Use Typst to compile
            typst_arguments = ['typst', 'compile', f'{outputfilebase}.typ']
            process = subprocess.run(typst_arguments,
                                    input=output,
                                    universal_newlines=True)
            log.info("PDF file has been compiled.")

        return output



from report_ranger.errors import InputError
import logging

log = logging.getLogger(__name__)


def combine_matrix(matrix, next_matrix):
    """ Combine two risk matrices where the final risk in one is fed into the next. It is called recursively """

    # Create a shallow copy of the matrix so we're not changing what is being passed to the function
    newmatrix = matrix.copy()

    for stage in newmatrix.keys():
        # See if we're at the end, otherwise recurse
        if isinstance(newmatrix[stage], dict):
            # We're in a dict so recurse
            newmatrix[stage] = combine_matrix(newmatrix[stage], next_matrix)
        else:
            # This is a reference to the next matrix so add it in
            if newmatrix[stage] not in next_matrix:
                log.error(
                    "Risk matrix validation error: In validating the risk assessment, we have not been able to find {} in the next matrix stage. Options are {}.".format(
                        newmatrix[stage], next_matrix.keys()))
                raise Exception("In validating the risk assessment, we have not been able to find {} in the next matrix stage. Options are {}.".format(
                    newmatrix[stage], next_matrix.keys()))
            else:
                # Add what's in the next matrix so this field is combined
                newmatrix[stage] = next_matrix[newmatrix[stage]]

    return newmatrix


def validate_matrix(matrix, stages, risks):
    """ Validates that each combination of stages is in the final matrix and the risks all make sense """
    if len(stages) == 0:
        log.error(
            "Risk matrix validation error: Excessively deep risk matrix found. We have the following part of the matrix, but we're out of risk stages. {}".format(
                matrix))
        # We've gotten to the end of stages, but we still have a matrix. This isn't good.
        raise Exception("Excessively deep risk matrix found. We have the following part of the matrix, but we're out of risk stages. {}".format(
            matrix))

    current_stage = stages[0]
    future_stages = stages[1:]

    for stage in current_stage['ratings']:
        if stage not in matrix:
            log.error(
                "Risk matrix validation error: Stage {} not found in the remaining matrix. All we have is {}".format(
                    stage, matrix.keys()))
            raise Exception("Stage {} not found in the remaining matrix. All we have is {}".format(
                stage, matrix.keys()))

        elif len(future_stages) == 0:
            if matrix[stage] not in risks:
                log.error(
                    "Risk matrix validation error: Could not find {} in the risks. The risks we have are {}".format(
                        matrix[stage], risks))
                raise Exception("Could not find {} in the risks. The risks we have are {}".format(
                    matrix[stage], risks))
            else:
                # Validate the next level of the matrix
                return validate_matrix(matrix[stage], future_stages, risks)


def get_risk(matrix, stages, headers):
    """
    A recursive assisting function that goes through the risk assessment tree to get the final risk
    """
    if len(stages) == 0:
        return matrix

    # Make sure that the next stage is actually in the headers
    if stages[0]['id'] not in headers:
        log.warn("{} not found in the headers of the vulnerability file as required by the template. Assigning risk as None.".format(
            stages[0]['id']))
        return "None"

    stageval = headers[stages[0]['id']]

    # If the header for this stage is a list of length one then take that value
    if type(stageval) is list:
        if len(stageval) != 1:
            stageval = stageval[0]
        else:
            log.warn("{} in this vulnerability was a list rather than a string. Assigning risk as None.".format(
                stages[0]['id']))
            return "None"

    # If the value of the header isn't in the matrix, cast an alert
    if stageval not in matrix:
        log.warn("'{}' is not a valid value for {}. The selection is: {}".format(
            stageval, stages[0]['id'], stages[0]['ratings']))
        return "None"

    return get_risk(matrix[headers[stages[0]['id']]], stages[1:], headers)


class RiskAssessment:
    """
    Holds the risk assessment, allows easy application of the risk assessment
    Can be customised
    """

    def __init__(self, riskassessment):
        self.name = riskassessment['name']
        self.id = riskassessment['id']
        self.stages = riskassessment['stages']
        self.risks = ["None", "Closed"] + riskassessment['risks']
        self.methodology = riskassessment['methodology']
        if 'mapping' in riskassessment:
            self.mapping = riskassessment['mapping']
        else:
            self.mapping = {}

        if 'style_text' in riskassessment:
            self.style_text = riskassessment['style_text']
        else:
            self.style_text = {}

        matrix = riskassessment['matrix']

        # First we see if there's a split matrix, and if so combine them into one
        if isinstance(matrix, list):
            # We have a split matrix. Combine them together
            combinedmatrix = matrix[0]
            for next_matrix in matrix[1:]:
                combinedmatrix = combine_matrix(combinedmatrix, next_matrix)

            self.matrix = combinedmatrix
        else:
            self.matrix = matrix

        # Validate the matrix
        validate_matrix(self.matrix, self.stages, self.risks)

        # Generate the table
        self.generate_riskmatrixtable()

    def get_risk(self, headers):
        """ Get the risk from the submitted risk assessment """

        if 'status' in headers and headers['status'] == 'Closed':
            return 'Closed'

        # Process the mappings
        for value, mapping in self.mapping.items():
            # We have to make sure we're not overwriting a value here
            if mapping not in headers and value in headers:
                headers[mapping] = headers[value]

        for stage in self.stages:
            # Make sure that the next stage is actually in the headers
            if stage['id'] not in headers:
                log.warn("{} not found in the headers of the vulnerability file as required by the template. Assigning risk as None.".format(
                    stage['id']))
                # Reset all stages to None if they're not there
                for s in self.stages:
                    if s['id'] not in headers:
                        headers[s['id']] = "None"
                return "None"

            if 'mapping' in stage:
                # Process the mappings
                for value, mapping in stage['mapping'].items():
                    if headers[stage['id']] == value:
                        headers[mapping] = headers[value]

        return get_risk(self.matrix, self.stages, headers)

    def generate_riskmatrixtable(self):
        """
        This displays a two stage risk matrix in a table. Note that it will only generate output if there are specifically 2 stages.
        """
        if len(self.stages) != 2:
            log.info(
                "Not auto generating a risk assessment table as there are not specifically 2 stages.")
            return ''

        sa = self.stages[0]
        sb = self.stages[1]
        theaders = [['thead'] * (len(sa['ratings']) + 1)]
        ratable = [[[sb['name'], sa['name']]] + sa['ratings']]
        sbratings = sb['ratings']
        sbratings.reverse()
        for sbrating in sbratings:
            rs = [sbrating]
            th = ['thead']
            for sarating in sa['ratings']:
                rs.extend(
                    [self.get_risk({sa['id']: sarating, sb['id']: sbrating})])
                th.extend(
                    ['t' + self.get_risk({sa['id']: sarating, sb['id']: sbrating}).lower().replace(' ', '')])
            ratable += [rs]
            theaders += [th]

        self.riskmatrixtable = ratable



from report_ranger import validation
from report_ranger.output_formatter.typstformatter import TypstFormatter
from report_ranger.utils.jinja_helpers import log_jinja2_error
from report_ranger.utils.mdread import markdown_from_file, process_included_header, process_template, _process_imports
from report_ranger.output_formatter.outputformatter import OutputFormatter
from report_ranger.riskassessment import RiskAssessment
from report_ranger.template import Template, retrieve_template
import os
import jinja2
import datetime
import logging
import traceback
import copy
import time
from report_ranger.output_formatter.latexformatter import LatexFormatter
from report_ranger.output_formatter.htmlformatter import HTMLFormatter
from report_ranger.output_formatter.csvformatter import CSVFormatter
from report_ranger.output_formatter.docxformatter import DOCXFormatter
import time

log = logging.getLogger(__name__)

def format_string_to_te(formatstring):
    if formatstring == "pdf" or formatstring == "pdf-typst":
        return "typst", "pdf"
    if formatstring == "pdf-latex":
        return "latex", "pdf"
    if formatstring == "markdown-latex":
        return "latex", "md"
    if formatstring == "typst":
        return "typst", "typ"
    if formatstring == "markdown":
        return "markdown", "md"
    if formatstring == "docx":
        return "docx", "docx"
    if formatstring == "html":
        return "html", "html"
    if formatstring == "csv":
        return "csv", "csv"
    return None, None

def get_format(commandline = None, output_file = None, config = None, reportheaders = None):
    log.debug(f"Getting format based on commandline: '{commandline}', output_file: '{output_file}', config: '{config}', reportheaders: '{reportheaders}'")
    # The order of preferences is commandline first, then the extension of the output file, then report headers, then config
    if commandline:
        target, ext = format_string_to_te(commandline)
        if target:
            log.info(f"Setting target to {target} and extension to {ext} based on command line args")
            return target, ext
    
    # Get the filename extension
    if output_file:
        fn, ext = os.path.splitext(output_file)
        if ext:
            ext = ext[1:]
            log.debug(f"Extension of output file is {ext}")
            if ext in ["docx", "html", "csv"]:
                log.info(f"Setting target to {ext} and extension to {ext} based on file extension")
                return ext, ext
            
            # For PDF and Markdown we need to know if there's a more specific target in the config or report headers. If so we use that.
            if ext in ["md", "pdf"]:
                for fs in [config, reportheaders]:
                    target, newext = format_string_to_te(fs)
                    if newext == ext:
                        log.info(f"Setting target to {target} and extension to {ext} based on file extension")
                        return target, ext
                return "markdown", "md" if ext == "md" else "typst", "pdf"

            if ext == "typ":
                log.info(f"Setting target to typst and extension to typ based on file extension")
                return "typst", "typ"
                
    # Now try config and report headers
    for format in [config, reportheaders]:
        target, ext = format_string_to_te(format)
        if target:
            return target, ext
    
    log.info("Setting default target to markdown and extension to md")
    return "markdown", "md"

def retrieve_output_formatter(target, templateheaders, timer, watcher=None):
    # Set up the appropriate output formatter
    if target == 'latex':
        return LatexFormatter(templateheaders, timer, watcher=watcher)
    elif target == 'html':
        return HTMLFormatter(templateheaders, timer, watcher=watcher)
    elif target == 'csv':
        of = CSVFormatter(templateheaders, timer, watcher=watcher)
    elif target == 'docx':
        of = DOCXFormatter(templateheaders, timer, watcher=watcher)
    elif target == 'typst':
        of = TypstFormatter(templateheaders, timer, watcher=watcher)
    else:  # Markdown formatter as default
        of = OutputFormatter(templateheaders, timer, watcher=watcher)
    return of

class Timer:
    def __init__(self):
        self._start_time=time.perf_counter()
    
    def time(self):
        return time.perf_counter() - self._start_time

class Report:
    """ Build a report from a collection of markdown. """

    def __init__(self, report_file, template:Template=None, templatefile:str="", default_template:str="", templatemapper:dict={}):
        self.timer = Timer()
        log.debug(f"Retrieving file markdown")
        self.report_file = report_file
        self.template = template
        self.templatefile = templatefile
        self.default_template = default_template
        self.templatemapper = templatemapper

        log.debug(f"Finished report init")

    # If the "change_template" header is there the report can set its own template variables
    def _change_template_headers(self, templateheaders, reportheaders):
        if 'change_template' in reportheaders:
            log.debug("change_template header detected")
            if not isinstance(reportheaders['change_template'], dict):
                log.warn(
                    "change_template front matter variable not a dictionary")
                return templateheaders

            # Notice this is back to front, the template itself is included into the report change_template variable
            templateheaders = process_included_header(
                reportheaders['change_template'], templateheaders)

        return templateheaders

    def attach_template(self, template:Template):
        self.template = template

    def process_file(self, mdfile, format=None, default_format=None, output=None, default_output_file=None, watcher=None):
        """ Process a file with the template. Returns the processed markdown for the file.

        The format should be one of ("latex", "pdf-latex", "pdf-typst", "docx", "typst")
        """

        
        log.debug(f"Begin processing file")
        if watcher:
            watcher.clear_paths()
            watcher.set_running()

        self.headers, self.markdown = markdown_from_file(self.report_file, process_imports=False, watcher=watcher)
        self.parentdir = os.path.dirname(os.path.join(os.path.curdir, self.report_file))
        self.templatefile = self.templatefile

        # Figure out where to get the template. In order of preference we take:
        # - Template provided in `template`
        # - Template provided in `templatefile`
        # - Template provided in the report, using templatemapper if necessary

        retrieved_template = False

        if self.template:
            # Template provided in `template`
            log.info("Using user provided template")
            self.template = self.template
            retrieved_template = True
        
        if not retrieved_template and self.templatefile != '':
            # Use the supplied template file
            try:
                self.template = retrieve_template(self.templatefile, self.templatemapper)
                log.info("Adding supplied template file")
                retrieved_template = True
            except Exception as e:
                log.warning(f"Template file provided cannot be loaded ({e.args}): {self.templatefile}")
        
        if not retrieved_template and 'template' in self.headers:
            # The template is specified in the report headers
            try:
                self.template = retrieve_template(self.headers['template'], self.templatemapper, self.parentdir)
                log.info("Adding template in report headers template file")
                retrieved_template = True
            except Exception as e:
                log.warning(f"Template in report headers cannot be loaded ({e.args}): {self.templatefile}")

        if not retrieved_template and self.default_template != "":
            # We're trying to fall back to the default template
            try:
                self.template = retrieve_template(self.default_template, self.templatemapper)
                log.info("Adding default template")
                retrieved_template = True
            except Exception as e:
                log.warning(f"Default template cannot be loaded ({e.args}): {self.default_template}")

        if self.template == None:
            log.warn("Could not figure out a template. Please set a template with the '-t' flag.")
            log.warn("Alternatively, use a config file. If you want to know where config files can be then use the '-v' flag and it will tell you.")
            log.warn("Configuration files are YAML, so you need to put in `template: template_location` in the config file.")
            log.warn(f"Default template: {self.default_template}")
            log.warn(f"Template mapper: {self.templatemapper}")
            raise Exception("Could not find template")
        

        # Make a new copy of the attached template
        templateheaders = copy.deepcopy(self.template.templateheaders)
        templatemarkdown = copy.deepcopy(self.template.templatemarkdown)

        templateheaders = self._change_template_headers(templateheaders, self.headers)

        # Initialise the risk assessment
        riskassessment = RiskAssessment(templateheaders['riskassessment'])
        if 'style_text' in templateheaders['riskassessment']:
            templateheaders['defaults']['ra_style_text'] = templateheaders['riskassessment']['style_text']

    
        log.debug(f"{self.timer.time()}: Creating output formatter")

        # Let's get the target and extension. First we need to make sure we've got appropriate output files
        output_file = output if output else self.headers.get('output_file')
        reportheaders_format = self.headers.get('output_format')
        target, ext = get_format(commandline=format, output_file = output_file, config=default_format, reportheaders=reportheaders_format)

        of = retrieve_output_formatter(target, templateheaders, self.timer, watcher=watcher)

        # Set up environment variables
        of.env.set_static('ra', riskassessment)
        of.env.set_static('templatedir', templateheaders['templatedir'])

        # The title defaults to the filename without the extension
        if 'title' not in self.headers:
            log.warn("There's no title set in the front matter, the default title is the filename of {}".format(
                os.path.basename(mdfile)))
            of.env.set_variable('title', os.path.basename(mdfile))

        # Convert date to the date object
        if 'date' not in self.headers:
            # The date defaults to the last change, otherwise today
            if 'changes' in self.headers and len(self.headers['changes']) != 0:
                of.env.set_variable('date', self.headers['changes'][-1][1])
                log.info(f"As there's no date in front matter, date of the latest change is used: {of.env.get('date')}")
            else:
                of.env.set_variable('date', datetime.date.today())
                log.info(f"As there's no date in front matter or changes, date has been set to today: {of.env.get('date')}")

        if 'client' not in self.headers:
            log.warning(
                "Client name is not in the report front matter. Default of [client] will be used.")
            of.env.set_variable('client', '[client]')

        if 'version' not in self.headers:
            if 'changes' in self.headers:
                of.env.set_variable('version', self.headers['changes'][-1][0])
            else:
                of.env.set_variable('version', '1.0')
                log.warning(
                    "Version is not in the report front matter. Default of 1.0 will be used.")
        
        # We now need to process imports. This will import sections, vulnerabilities, XLSX and CSV files
        # We can't process imports before because the output formatter wasn't set up yet
        of.env.set_variables(self.headers)
        
        log.debug("Setting reportbody imports")
        if 'defaults' in templateheaders:
            template_import_headers = _process_imports(
                templateheaders['defaults'],
                of.env,
                watcher=watcher,
                pathlist=[self.template.templatedir]
                )
            of.env.set_variables(template_import_headers)
        import_headers = _process_imports(
            self.headers,
            of.env,
            watcher=watcher
            )
        of.env.set_variables(import_headers)
        
        # Make sure that the risk assessment methodology is processed as RR markdown
        riskassessment.methodology_markdown = process_template(
            {}, riskassessment.methodology, env=of.env, name="Risk assessment methodology")

        if 'validation' in templateheaders:
            if 'report' in templateheaders['validation']:
                validation.validate_headers(
                    templateheaders['validation']['report'], self.headers, validation.default_report_validation)

        if output == None:
            if of.env.get('output_file'):
                rawfilename = of.env.get('output_file')
                try:
                    output = process_template({}, rawfilename, of.env, 'output_file')
                except Exception as error:
                    log.warn("Error processing filename header: {}".format(error.args))
                    output = f"{os.path.basename(mdfile)}.{ext}"
                    log.warn(f"Setting to {output}")
            elif default_output_file:
                log.info(f"Setting to default output file {default_output_file}")
                output = default_output_file
            else:
                output = f"{os.path.basename(mdfile)}.{ext}"
                log.warn(f"Filename not found, setting to {output}")
        
        # Let's make sure the output file has the extension, if not add it
        
        fn, ofext = os.path.splitext(output)
        if ofext != "." + ext:
            output += "." + ext

        log.debug(f"{self.timer.time()}: Begin processing report body")

        # Render reportbody
        try:
            j2template = jinja2.Template(self.markdown)
            rbrendered = j2template.render(of.env.get_env())
            of.env.set_static('reportbody', rbrendered)
        except jinja2.exceptions.TemplateSyntaxError as error:
            log.error("Report body render Jinja2 error: {} at lineno {} in reportbody for file {}".format(
                error.message, error.lineno, error.filename))
            log_jinja2_error(self.markdown, error)
            of.env.set_static('reportbody', "")
        except Exception as error:
            log.error("Exception found in reportbody: {}".format(error.args))
            traceback.print_exc()
            log.error("Removing reportbody text")
            of.env.set_static('reportbody', "")

        of.docformat = ext
        of.options = {}
        
        log.debug(f"Begin outputting")
        
        if watcher:
            watcher.stop_running()

        return of.output(templatemarkdown, output)



# from .mdread import markdown_from_directory, process_template
from report_ranger.utils import mdread
import logging
import traceback
import jinja2

from report_ranger.utils.jinja_helpers import log_jinja2_error

log = logging.getLogger(__name__)


def int_to_rn(num, uppercase=True):
    val = [1000, 900, 500, 400, 100, 90, 50, 40, 10, 9, 5, 4, 1]
    uppercase = ["M", "CM", "D", "CD", "C", "XC",
                 "L", "XL", "X", "IX", "V", "IV", "I"]
    lowercase = ["m", "cm", "d", "cd", "c", "xc",
                 "l", "xl", "x", "ix", "v", "iv", "i"]
    numeral = ''
    i = 0
    while num > 0:
        for _ in range(num // val[i]):
            if uppercase:
                numeral += uppercase[i]
            else:
                numeral += lowercase[i]
            num -= val[i]
        i += 1
    return numeral


class Section:
    def __init__(self, directory, ordinal='1', watcher=None):
        try:
            chaptersmd = mdread.markdown_from_directory(directory, watcher=watcher)
            if watcher:
                watcher.add_path(directory)
        except FileNotFoundError as e:
            log.warn(e.args[0])
            self.chapters = []
            return
        except Exception as e:
            log.error(
                "Received exception when processing directory {}: {}".format(directory, e.args))
            # log.error(traceback.format_exc()) Not displaying traceback because it's usually just a missing directory
            self.chapters = []
            return

        # Sort the chapters by weight
        for i in chaptersmd:
            if 'weight' not in i[0]:
                i[0]['weight'] = 1
        chaptersmd = sorted(
            chaptersmd, key=lambda x: float(x[0]['weight']) * -1)

        nchapter = 0
        chapters = []

        for i in range(len(chaptersmd)):
            if ordinal == 'a':
                ref = chr(ord('a') + nchapter)
            elif ordinal == 'A':
                ref = chr(ord('A') + nchapter)
            elif ordinal == 'I':
                ref = int_to_rn(nchapter + 1)
            elif ordinal == 'i':
                ref = int_to_rn(nchapter + 1, False)
            else:
                ref = nchapter + 1

            chapters.append(
                Chapter(ref, chaptersmd[i]))
            nchapter += 1

        self.chapters = chapters

    def get_chapters(self, env):
        chapters = []
        for chapter in self.chapters:
            chapters.append(chapter.get_chapter(env))
        return chapters


class Chapter:
    def __init__(self, ref, chaptermd, filename="name"):
        self.headers = chaptermd[0]
        self.headers['ref'] = ref
        self.chapter_markdown = chaptermd[1]
        self.ref = ref
        self.filename = self.headers['filename']
        self.filename_noext = self.headers['filename_noext']

        if 'name' not in self.headers:
            if 'Name' in self.headers:
                self.headers['name'] = self.headers['Name']
            else:
                log.warn(f"Chapter from file '{self.filename}' does not have a name. Setting it to '{self.filename_noext}''")
                self.headers['name'] = 'Chapter'

        self.name = self.headers['name']

    def markdown(self, env):
        try:
            output = mdread.process_template(
                self.headers, self.chapter_markdown, env)
            return output
        except jinja2.exceptions.TemplateSyntaxError as error:
            log.error("Jinja2 error processing chapter {}: {} at lineno {} for file {}".format(
                self.name, error.message, error.lineno, filename=error.filename))
            log.error("Removing markdown for chapter {}".format(self.name))
            log_jinja2_error(self.chapter_markdown, error)
            return ""
        except Exception as e:
            log.error(
                "Received exception when processing markdown for section {}: {}".format(self.name, e.args))
            log.error(traceback.format_exc())
            log.error(
                "Removing markdown for chapter {}".format(self.name))
            return ""

    def get_chapter(self, env):
        chapter = {}
        chapter['name'] = self.name
        chapter['headers'] = self.headers
        chapter['ref'] = self.ref
        chapter['markdown'] = self.markdown(env)
        return chapter



import time
from report_ranger.imports import vulnerability
from report_ranger.config import get_config  # Configuration settings within report ranger
from report_ranger.errors import InputError
from report_ranger.report import Report
from report_ranger.template import Template
from report_ranger.templatemapper import process_templatemapper
import os
import jinja2
import logging
from watchdog.observers import Observer
from report_ranger.watcher import Watcher
import argparse

from report_ranger.utils.jinja_helpers import log_jinja2_error

log = logging.getLogger(__name__)

version = "3.3.1"

def main():
    parser = argparse.ArgumentParser(
        description=f"Report Ranger collects together a bunch of markdown files to create a master markdown file that is suitable to use with Pandoc and Latex to build a PDF file. Version {version}.")
    parser.add_argument('-i', '--input', type=str,
                        help='The main report file to process')
    parser.add_argument('-o', '--output', type=str,
                        help='The file to output the final markdown to. If given a directory, the filename will be whatever is suggested by the report or template.')
    parser.add_argument('-f', '--format', type=str,
                        default='', help='The format to target (options are "latex", "pdf-latex", "pdf-typst", "docx", "typst"). Defaults to whatever the extension tells us.')
    parser.add_argument('-w', '--watch', action='store_true', default=False,
                        help='Watch files for changes and recompile if they are identified')
    parser.add_argument('--watch_mode', type=str, default='',
                        help='Watch file mode, can be set to "os" or "modified"')
    parser.add_argument('-m', '--templatemapper', type=str, default='',
                        help='A template mapper file which holds a YAML mapping from a template name to a related file.')
    parser.add_argument('-c', '--config', type=str, default='',
                        help='The location of a config file in YAML format.')
    parser.add_argument('-t', '--template', type=str, default='',
                        help='The template file. Associated images should be in the same directory. Defaults to what is set in the report.')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help=f'Turn on verbose mode.')
    parser.add_argument('--version', action='store_true',
                        help=f'Print the version and exit.')

    args = parser.parse_args()

    if args.version:
        print(f"Report Ranger version {version}")
        return

    # Turn on verbose mode
    if args.verbose:
        logging.basicConfig(
            format='%(levelname)s: %(message)s', level=logging.INFO)
    else:
        logging.basicConfig(format='%(levelname)s: %(message)s',
                            level=logging.WARNING)

    config = get_config(args.config)

    input = args.input if args.input else config['default_input_file'] if 'default_input_file' in config else None
    if input == None or input == '':
        log.warn("Report Ranger needs a file as input. Please use the -i flag to set an input file.")
        return
    output_file = args.output if args.output else None

    parentdir = os.path.dirname(os.path.join(os.path.curdir, input))

    # Sort out command line format
    default_format = config['default_format'] if 'default_format' in config else None
    format = args.format if args.format else None

    # We need to change the current working directory to the directory of the template otherwise relative
    # paths inside the template won't work. For instance you won't be able to include executivesummary.md
    rr_parent_folder = os.path.abspath(os.path.curdir)

    ctm = config['templatemapper'] if 'templatemapper' in config else {}        
    ctms = config['templatemappers'] if 'templatemappers' in config else []

    templatemapper = process_templatemapper(args.templatemapper, ctm, ctms)

    if args.template:
        if args.template in templatemapper:
            templatefile = templatemapper[args.template]
        else:
            templatefile = os.path.abspath(args.template)
    else:
        templatefile = ''

    os.chdir(parentdir)
    parentdir = '.'
    mdfile = os.path.basename(input)

    # Initialise the report
    try:
        report = Report(
            mdfile,
            templatefile=templatefile,
            templatemapper=templatemapper,
            default_template=config['default_template'])
    except Exception as e:
        log.warn(f"Could not initialise report: {e.args}")
        return
    except:
        log.warn("Could not initialise report.")
        return

    # Pandoc does not support PDF output to stdout, so we need to hack it by
    # making a symlink to /dev/stdout and outputting to that file
    stdout_link = None
    if format == 'pdf' and output_file == '-':
        stdout_link = '/tmp/stdout.pdf'
        os.symlink('/dev/stdout', stdout_link)
        output_file = stdout_link

    # Convert output file path into full path if relative path is given
    if output_file and output_file[0] != '/':
        output_file = os.path.join(rr_parent_folder, output_file)

    if args.watch:
        log.debug("Starting watch")
        try:
            watcher = Watcher(log.debug, "Callback")
            watcher.set_callback(report.process_file, mdfile, format=format, default_format=default_format, output = output_file, default_output_file=config.get('default_output_file'), watcher=watcher)
            watcher.set_watch_mode(args.watch_mode)
            output = report.process_file(mdfile, format=format, default_format=default_format, output = output_file, default_output_file=config.get('default_output_file'), watcher=watcher)
            if args.watch_mode != "modified":
                log.info("Setting watch mode to os")
                observer = Observer()
                observer.schedule(watcher, parentdir, recursive=True)
                observer.start()
                try:
                    while True:
                        time.sleep(5)
                        watcher.run()
                finally:
                    observer.stop()
                    observer.join()
            else:
                log.info("Setting watch mode to modification time")
                while True:
                    try:
                        time.sleep(5)
                        watcher.run()
                    except InputError as ie:
                        log.error("Input Error: {}".format(ie.message))
                        exit()
                    except jinja2.exceptions.TemplateSyntaxError as error:
                        log.error("Final report processing Jinja2 error: {} at lineno {} for file {}".format(
                            error.message, error.lineno, error.filename))
                        log_jinja2_error(mdfile, error)
                        exit()

        except InputError as ie:
            log.error("Input Error: {}".format(ie.message))
        except jinja2.exceptions.TemplateSyntaxError as error:
            log.error("Final report processing Jinja2 error: {} at lineno {} for file {}".format(
                error.message, error.lineno, error.filename))
            log_jinja2_error(mdfile, error)
    else:
        try:
            output = report.process_file(mdfile, format=format, default_format=default_format, output = output_file, default_output_file=config.get('default_output_file'))
        except InputError as ie:
            log.error("Input Error: {}".format(ie.message))
            exit()
        except jinja2.exceptions.TemplateSyntaxError as error:
            log.error("Final report processing Jinja2 error: {} at lineno {} for file {}".format(
                error.message, error.lineno, error.filename))
            log_jinja2_error(mdfile, error)
            exit()

        # If we're outputting to stdout, remove the link
        if stdout_link and os.path.exists(stdout_link):
            os.remove(stdout_link)



import mistune
import re


class CSVRenderer(mistune.HTMLRenderer):
    NAME = 'csv'
    IS_TREE = False

    # This set of data structures is a bit 'special'.
    #
    # We want the template to be able to refer to headings and get all the text from under that header. Usually you could do this with a tree, but
    # I chose a different way here. Instead of a tree, it has a dict and a list. The dict gives the headings from every level and the list gives the current
    # headings in play. When the renderer gets a new piece of text, it adds it to all the headings currently in play. This means it's very easy to get
    # and refer to all the text from the headings without having to traverse a tree.

    def _add_text(self, text):

        for heading in self.heading_levels:
            if heading != '' and heading in self.heading_text:
                self.heading_text[heading] += text

    def heading(self, text, level):
        # This is brought to the top as it relates to the above heading structures

        if len(self.heading_levels) < level:
            self.heading_levels += [''] * \
                (level - len(self.heading_levels)) + [text]
        else:
            self.heading_levels = self.heading_levels[:level - 1] + [text]

        for heading in self.heading_levels[-1:]:
            if heading != '' and heading in self.heading_text:
                self.heading_text[heading] += '#' * level + ' ' + text

        self.heading_text[text] = ''

        return '#' * level + ' ' + text

    def __init__(self, escape=True):
        self.heading_levels = []  # The current headings in play
        self.heading_text = {}  # The current heading text
        super(CSVRenderer, self).__init__()

    def text(self, text):
        return text

    def link(self, link, text=None, title=None):
        self._add_text(text)
        return text

    def image(self, src, alt="", title=None):
        self._add_text(alt)
        return alt

    def emphasis(self, text):
        self._add_text(text)
        return text

    def strong(self, text):
        self._add_text(text)
        return text

    def codespan(self, text):
        self._add_text(text)
        return text

    def linebreak(self):
        self._add_text('\n\n')
        return '\n\n'

    def inline_html(self, html):
        self._add_text(self._escape_csv(html))
        return self._escape_csv(html)

    def paragraph(self, text):
        self._add_text(text + '\n\n')
        return text + '\n\n'

    def newline(self):
        self._add_text('\n\n')
        return '\n\n'

    def thematic_break(self):
        # TODO NO THEMATIC BREAK
        return '\n\n'

    def block_text(self, text):
        return text

    def block_code(self, code, info=None):
        return code

    def block_quote(self, text):
        return text

    def block_html(self, html):
        return html

    def block_error(self, html):
        return html

    def list(self, text, ordered, level, start=None):
        return text

    def list_item(self, text, level):
        return '  ' * level + '- ' + text + '\n'
